# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import cast
from xml.etree import ElementTree

from typing_extensions import override

from .decodestate import DecodeState
from .diagcodedtype import DctType, DiagCodedType
from .encodestate import EncodeState
from .encoding import get_string_encoding
from .exceptions import DecodeError, EncodeError, odxassert, odxraise, odxrequire
from .odxdoccontext import OdxDocContext
from .odxtypes import AtomicOdxType, BytesTypes, DataType
from .termination import Termination
from .utils import dataclass_fields_asdict


@dataclass(kw_only=True)
class MinMaxLengthType(DiagCodedType):
    max_length: int | None = None
    min_length: int
    termination: Termination

    @property
    def dct_type(self) -> DctType:
        return "MIN-MAX-LENGTH-TYPE"

    @staticmethod
    @override
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "MinMaxLengthType":
        kwargs = dataclass_fields_asdict(DiagCodedType.from_et(et_element, context))

        max_length = None
        if et_element.find("MAX-LENGTH") is not None:
            max_length = int(odxrequire(et_element.findtext("MAX-LENGTH")))
        min_length = int(odxrequire(et_element.findtext("MIN-LENGTH")))

        termination_str = odxrequire(et_element.attrib.get("TERMINATION"))
        try:
            termination = Termination(termination_str)
        except ValueError:
            termination = cast(Termination, None)
            odxraise(f"Encountered unknown termination type '{termination_str}'")

        return MinMaxLengthType(
            max_length=max_length, min_length=min_length, termination=termination, **kwargs)

    def __post_init__(self) -> None:
        odxassert(self.max_length is None or self.min_length <= self.max_length)
        odxassert(
            self.base_data_type in [
                DataType.A_BYTEFIELD,
                DataType.A_ASCIISTRING,
                DataType.A_UNICODE2STRING,
                DataType.A_UTF8STRING,
            ], f"A min-max length type cannot have the base data type {self.base_data_type}.")

    def __termination_sequence(self) -> bytes:
        """Returns the termination byte sequence if it isn't defined."""
        # The termination sequence is actually not specified by ASAM
        # for A_BYTEFIELD but I assume it is only one byte.
        termination_sequence = b''
        if self.termination == Termination.ZERO:
            if self.base_data_type not in [DataType.A_UNICODE2STRING]:
                termination_sequence = bytes([0x0])
            else:
                termination_sequence = bytes([0x0, 0x0])
        elif self.termination == Termination.HEX_FF:
            if self.base_data_type not in [DataType.A_UNICODE2STRING]:
                termination_sequence = bytes([0xFF])
            else:
                termination_sequence = bytes([0xFF, 0xFF])
        return termination_sequence

    @override
    def encode_into_pdu(self, internal_value: AtomicOdxType, encode_state: EncodeState) -> None:

        if not isinstance(internal_value, (str, BytesTypes)):
            odxraise("MinMaxLengthType is currently only implemented for strings and byte arrays",
                     EncodeError)

        raw_value = b''
        if isinstance(internal_value, str):
            str_encoding = get_string_encoding(self.base_data_type, self.base_type_encoding,
                                               self.is_highlow_byte_order)

            if str_encoding is None:
                odxraise(f"Internal string value specified for object which is "
                         f"'{self.base_data_type.value}' not a string")
                raw_value = b''
            else:
                raw_value = internal_value.encode(str_encoding)
        else:
            raw_value = bytes(internal_value)

        data_length = len(raw_value)

        if data_length < self.min_length:
            odxraise(
                f"Encoded value for MinMaxLengthType "
                f"must be at least {self.min_length} bytes long. "
                f"(Is: {data_length} bytes.)", EncodeError)
            data_length = self.min_length
        elif self.max_length is not None and data_length > self.max_length:
            odxraise(
                f"Encoded value for MinMaxLengthType "
                f"must not be longer than {self.max_length} bytes. "
                f"(Is: {data_length} bytes.)", EncodeError)
            data_length = self.max_length

        encode_state.emplace_atomic_value(
            internal_value=raw_value,
            used_mask=None,
            bit_length=8 * data_length,
            base_data_type=DataType.A_BYTEFIELD,
            base_type_encoding=None,
            is_highlow_byte_order=True,
        )

        # TODO: ensure that the termination delimiter is not
        # encountered within the encoded value.

        odxassert(
            self.termination != Termination.END_OF_PDU or encode_state.is_end_of_pdu,
            "Encountered a MIN-MAX-LENGTH type with END-OF-PDU termination "
            "which is not located at the end of the PDU")
        if encode_state.is_end_of_pdu or data_length == self.max_length:
            # All termination types may be ended by the end of the PDU
            # or once reaching the maximum length. In this case, we
            # must not add the termination sequence
            pass
        else:
            termination_sequence = self.__termination_sequence()

            # ensure that we don't try to encode an odd-length
            # value when using a two-byte terminator
            odxassert(data_length % len(termination_sequence) == 0)

            data_length += len(termination_sequence)
            encode_state.emplace_bytes(termination_sequence)

    def decode_from_pdu(self, decode_state: DecodeState) -> AtomicOdxType:
        odxassert(decode_state.cursor_bit_position == 0,
                  "No bit position can be specified for MIN-MAX-LENGTH-TYPE values.")
        if decode_state.cursor_byte_position + self.min_length > len(decode_state.coded_message):
            raise DecodeError("The PDU ended before minimum length was reached.")

        coded_message = decode_state.coded_message
        orig_cursor_pos = decode_state.cursor_byte_position
        termination_seq = self.__termination_sequence()

        max_terminator_pos = len(coded_message)
        if self.max_length is not None:
            max_terminator_pos = min(max_terminator_pos, orig_cursor_pos + self.max_length)

        if self.termination != Termination.END_OF_PDU:
            # The parameter either ends after the maximum length, at
            # the end of the PDU or if a termination sequence is
            # found.

            # Find the location of the termination sequence. The
            # problem here is that the alignment of the termination
            # sequence must be correct for it to be a termination
            # sequence. (e.g., an odd-aligned double-zero for UTF-16
            # strings is *not* a termination sequence!)
            terminator_pos = orig_cursor_pos + self.min_length
            while True:
                terminator_pos = decode_state.coded_message.find(termination_seq, terminator_pos,
                                                                 max_terminator_pos)
                if terminator_pos < 0:
                    # termination sequence was not found, i.e., we
                    # are terminated by either the end of the PDU or
                    # our maximum size. (whatever is the smaller
                    # value.)
                    byte_length = max_terminator_pos - orig_cursor_pos
                    break
                elif (terminator_pos - orig_cursor_pos) % len(termination_seq) == 0:
                    # we found the termination sequence at a position
                    # and it is correctly aligned (two-byte
                    # termination sequences must be word aligned
                    # relative to the beginning of the parameter)!
                    byte_length = terminator_pos - orig_cursor_pos
                    break
                else:
                    # we found the termination sequence, but its
                    # alignment was incorrect. Try again one byte
                    # further...
                    terminator_pos += 1

            # Extract the value
            value = decode_state.extract_atomic_value(
                bit_length=8 * byte_length,
                base_data_type=self.base_data_type,
                base_type_encoding=self.base_type_encoding,
                is_highlow_byte_order=self.is_highlow_byte_order,
            )

            if decode_state.cursor_byte_position != len(
                    decode_state.coded_message
            ) and decode_state.cursor_byte_position - orig_cursor_pos != self.max_length:
                # next object starts after the actual data and the termination sequence
                decode_state.cursor_byte_position += len(termination_seq)

            return value
        else:
            # If termination == "END-OF-PDU", the parameter ends after max_length
            # or at the end of the PDU.
            byte_length = max_terminator_pos - orig_cursor_pos

            value = decode_state.extract_atomic_value(
                bit_length=8 * byte_length,
                base_data_type=self.base_data_type,
                base_type_encoding=self.base_type_encoding,
                is_highlow_byte_order=self.is_highlow_byte_order,
            )

            return value



# SPDX-License-Identifier: MIT
from collections.abc import Callable, Iterable
from copy import copy, deepcopy
from dataclasses import dataclass
from functools import cached_property
from itertools import chain
from typing import Any, Union, cast
from xml.etree import ElementTree

from ..admindata import AdminData
from ..companydata import CompanyData
from ..description import Description
from ..diagcomm import DiagComm
from ..diagdatadictionaryspec import DiagDataDictionarySpec
from ..diagservice import DiagService
from ..exceptions import DecodeError, odxassert, odxraise
from ..library import Library
from ..message import Message
from ..nameditemlist import NamedItemList, TNamed
from ..odxdoccontext import OdxDocContext
from ..odxlink import OdxLinkDatabase, OdxLinkId, OdxLinkRef
from ..parentref import ParentRef
from ..request import Request
from ..response import Response
from ..servicebinner import ServiceBinner
from ..singleecujob import SingleEcuJob
from ..snrefcontext import SnRefContext
from ..specialdatagroup import SpecialDataGroup
from ..subcomponent import SubComponent
from ..unitgroup import UnitGroup
from .diaglayerraw import DiagLayerRaw
from .diaglayertype import DiagLayerType

PrefixTree = dict[int, Union[list[DiagService], "PrefixTree"]]


@dataclass(kw_only=True)
class DiagLayer:
    """This class represents a "logical view" upon a diagnostic layer
    according to the ODX standard.

    i.e. it handles the value inheritance, communication parameters,
    encoding/decoding of data, etc.
    """

    diag_layer_raw: DiagLayerRaw

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "DiagLayer":
        diag_layer_raw = DiagLayerRaw.from_et(et_element, context)

        # Create DiagLayer
        return DiagLayer(diag_layer_raw=diag_layer_raw)

    def __post_init__(self) -> None:
        if self.diag_layer_raw.diag_data_dictionary_spec is None:
            # create an empry DiagDataDictionarySpec object if the raw
            # layer does not define a DDDS...
            self._diag_data_dictionary_spec = DiagDataDictionarySpec(
                admin_data=None,
                data_object_props=NamedItemList(),
                dtc_dops=NamedItemList(),
                structures=NamedItemList(),
                static_fields=NamedItemList(),
                end_of_pdu_fields=NamedItemList(),
                dynamic_endmarker_fields=NamedItemList(),
                dynamic_length_fields=NamedItemList(),
                tables=NamedItemList(),
                env_data_descs=NamedItemList(),
                env_datas=NamedItemList(),
                muxs=NamedItemList(),
                unit_spec=None,
                sdgs=[])
        else:
            self._diag_data_dictionary_spec = self.diag_layer_raw.diag_data_dictionary_spec

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        """Construct a mapping from IDs to all objects that are contained in this diagnostic layer."""
        result = self.diag_layer_raw._build_odxlinks()

        # we want to get the full diag layer, not just the raw layer
        # when referencing...
        result[self.odx_id] = self

        return result

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        """Recursively resolve all ODXLINK references."""

        # deal with the import references: these basically extend the
        # pool of objects that are referenceable without having to
        # explicitly specify the DOCREF attribute in the
        # reference. This mechanism can thus be seen as a kind of
        # "poor man's inheritance".
        if self.import_refs:
            imported_links: dict[OdxLinkId, Any] = {}
            for import_ref in self.import_refs:
                imported_dl = odxlinks.resolve(import_ref, DiagLayer)

                odxassert(
                    imported_dl.variant_type == DiagLayerType.ECU_SHARED_DATA,
                    f"Tried to import references from diagnostic layer "
                    f"'{imported_dl.short_name}' of type {imported_dl.variant_type.value}. "
                    f"Only ECU-SHARED-DATA layers may be referenced using the "
                    f"IMPORT-REF mechanism")

                # TODO: ensure that the imported diagnostic layer has
                # not been referenced in any PARENT-REF of the current
                # layer or any of its parents.

                # TODO: detect and complain about cyclic IMPORT-REFs

                # TODO (?): detect conflicts with locally-defined
                # objects

                imported_dl_links = imported_dl._build_odxlinks()
                for link_id, obj in imported_dl_links.items():
                    # the imported objects shall behave as if they
                    # were defined by the importing layer. IOW, they
                    # must be visible in the same document fragments.
                    link_id = OdxLinkId(link_id.local_id, self.odx_id.doc_fragments)
                    imported_links[link_id] = obj

            # We need to copy the odxlink database here since this
            # function must not modify its argument because the
            # imported references only apply within this specific
            # diagnostic layer
            extended_odxlinks = copy(odxlinks)
            extended_odxlinks.update(imported_links, overwrite=False)

            self.diag_layer_raw._resolve_odxlinks(extended_odxlinks)
            return

        self.diag_layer_raw._resolve_odxlinks(odxlinks)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        self.diag_layer_raw._resolve_snrefs(context)

    def _get_local_diag_comms(self, odxlinks: OdxLinkDatabase) -> Iterable[DiagComm]:
        """Return the list of locally defined diagnostic communications.

        This is not completely trivial as it requires to resolving the
        references specified in the <DIAG-COMMS> XML tag.
        """
        return self.diag_layer_raw.diag_comms

    def _get_local_unit_groups(self) -> Iterable[UnitGroup]:
        if self.diag_layer_raw.diag_data_dictionary_spec is None:
            return []

        unit_spec = self.diag_layer_raw.diag_data_dictionary_spec.unit_spec
        if unit_spec is None:
            return []

        return unit_spec.unit_groups

    def _compute_available_objects(
        self,
        get_local_objects: Callable[["DiagLayer"], Iterable[TNamed]],
        get_not_inherited: Callable[[ParentRef], Iterable[str]],
    ) -> Iterable[TNamed]:
        """Helper method to compute the set of all objects applicable
        to the DiagLayer if these objects are subject to the value
        inheritance mechanism

        This is the simplified version for diag layers which do not
        have parents and thus do not deal with value inheritance
        (i.e., ECU-SHARED-DATA).

        """
        return get_local_objects(self)

    def __deepcopy__(self, memo: dict[int, Any]) -> Any:
        """Create a deep copy of the diagnostic layer

        Note that the copied diagnostic layer is not fully
        initialized, so `_finalize_init()` should to be called on it
        before it can be used normally.
        """
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result

        result.diag_layer_raw = deepcopy(self.diag_layer_raw, memo)

        return result

    #####
    # <convenience functionality>
    #####
    @cached_property
    def service_groups(self) -> ServiceBinner:
        return ServiceBinner(self.services)

    #####
    # </convenience functionality>
    #####

    #####
    # <properties forwarded to the "raw" diag layer>
    #####
    @property
    def variant_type(self) -> DiagLayerType:
        return self.diag_layer_raw.variant_type

    @property
    def odx_id(self) -> OdxLinkId:
        return self.diag_layer_raw.odx_id

    @property
    def short_name(self) -> str:
        return self.diag_layer_raw.short_name

    @property
    def long_name(self) -> str | None:
        return self.diag_layer_raw.long_name

    @property
    def description(self) -> Description | None:
        return self.diag_layer_raw.description

    @property
    def admin_data(self) -> AdminData | None:
        return self.diag_layer_raw.admin_data

    @property
    def diag_comms(self) -> NamedItemList[DiagComm]:
        return self.diag_layer_raw.diag_comms

    @property
    def services(self) -> NamedItemList[DiagService]:
        return self.diag_layer_raw.services

    @property
    def diag_services(self) -> NamedItemList[DiagService]:
        return self.diag_layer_raw.diag_services

    @property
    def single_ecu_jobs(self) -> NamedItemList[SingleEcuJob]:
        return self.diag_layer_raw.single_ecu_jobs

    @property
    def company_datas(self) -> NamedItemList[CompanyData]:
        return self.diag_layer_raw.company_datas

    @property
    def requests(self) -> NamedItemList[Request]:
        return self.diag_layer_raw.requests

    @property
    def positive_responses(self) -> NamedItemList[Response]:
        return self.diag_layer_raw.positive_responses

    @property
    def negative_responses(self) -> NamedItemList[Response]:
        return self.diag_layer_raw.negative_responses

    @property
    def global_negative_responses(self) -> NamedItemList[Response]:
        return self.diag_layer_raw.global_negative_responses

    @property
    def import_refs(self) -> list[OdxLinkRef]:
        return self.diag_layer_raw.import_refs

    @property
    def libraries(self) -> NamedItemList[Library]:
        return self.diag_layer_raw.libraries

    @property
    def sub_components(self) -> NamedItemList[SubComponent]:
        return self.diag_layer_raw.sub_components

    @property
    def sdgs(self) -> list[SpecialDataGroup]:
        return self.diag_layer_raw.sdgs

    @property
    def diag_data_dictionary_spec(self) -> DiagDataDictionarySpec:
        """The DiagDataDictionarySpec applicable to this DiagLayer"""
        return self._diag_data_dictionary_spec

    #####
    # </properties forwarded to the "raw" diag layer>
    #####

    #####
    # <PDU decoding>
    #####
    @cached_property
    def _prefix_tree(self) -> PrefixTree:
        """Constructs the coded prefix tree of the services.

        Each leaf node is a list of `DiagService`s.  (This is because
        navigating from a service to the request/ responses is easier
        than finding the service for a given request/response object.)

        Example:
        Let there be four services with corresponding requests:
        * Request 1 has the coded constant prefix `12 34`.
        * Request 2 has the coded constant prefix `12 34`.
        * Request 3 has the coded constant prefix `12 56`.
        * Request 4 has the coded constant prefix `12 56 00`.

        Then, the constructed prefix tree is the dict
        ```
        {0x12: {0x34: {-1: [<Service 1>, <Service 2>]},
                0x56: {-1: [<Service 3>],
                       0x0: {-1: [<Service 4>]}
                       }}}
        ```
        Note, that the inner `-1` are constant to distinguish them
        from possible service IDs.

        Also note, that it is actually allowed that
        (a) SIDs for different services are the same like for service
            1 and 2 (thus each leaf node is a list) and
        (b) one SID is the prefix of another SID like for service 3
            and 4 (thus the constant `-1` key).

        """
        prefix_tree: PrefixTree = {}
        for s in self.services:
            # Compute prefixes for the service's request and all
            # possible responses. We need to consider the global
            # negative responses here, because they might contain
            # MATCHING-REQUEST parameters. If these global responses
            # do not contain such parameters, this will potentially
            # result in an enormous amount of decoded messages for
            # global negative responses. (I.e., one for each
            # service. This can be avoided by specifying the
            # corresponding request for `decode_response()`.)
            request_prefix = b''
            if s.request is not None:
                request_prefix = s.request.coded_const_prefix()
            prefixes = [request_prefix]
            gnrs = getattr(self, "global_negative_responses", [])
            prefixes += [
                x.coded_const_prefix(request_prefix=request_prefix)
                for x in chain(s.positive_responses, s.negative_responses, gnrs)
            ]
            for coded_prefix in prefixes:
                self._extend_prefix_tree(prefix_tree, coded_prefix, s)

        return prefix_tree

    @staticmethod
    def _extend_prefix_tree(prefix_tree: PrefixTree, coded_prefix: bytes,
                            service: DiagService) -> None:

        # make sure that tree has an entry for the given prefix
        sub_tree = prefix_tree
        for b in coded_prefix:
            if b not in sub_tree:
                sub_tree[b] = {}
            sub_tree = cast(PrefixTree, sub_tree[b])

        # Store the object as in the prefix tree. This is done by
        # assigning the list of possible objects to the key -1 of the
        # dictionary (this is quite hacky...)
        if sub_tree.get(-1) is None:
            sub_tree[-1] = [service]
        else:
            cast(list[DiagService], sub_tree[-1]).append(service)

    def _find_services_for_uds(self, message: bytes) -> list[DiagService]:
        prefix_tree = self._prefix_tree

        # Find matching service(s) in prefix tree
        possible_services: list[DiagService] = []
        for b in message:
            if b in prefix_tree:
                odxassert(isinstance(prefix_tree[b], dict))
                prefix_tree = cast(PrefixTree, prefix_tree[b])
            else:
                break
            if -1 in prefix_tree:
                possible_services += cast(list[DiagService], prefix_tree[-1])
        return possible_services

    def _decode(self, message: bytes, candidate_services: Iterable[DiagService]) -> list[Message]:
        decoded_messages: list[Message] = []

        for service in candidate_services:
            try:
                decoded_messages.append(service.decode_message(message))
            except DecodeError as e:
                # check if the message can be decoded as a global
                # negative response for the service
                gnr_found = False
                for gnr in self.global_negative_responses:
                    try:
                        decoded_gnr = gnr.decode(message)
                        gnr_found = True
                        if not isinstance(decoded_gnr, dict):
                            odxraise(
                                f"Expected the decoded value of a global "
                                f"negative response to be a dictionary, "
                                f"got {type(decoded_gnr)} for {self.short_name}", DecodeError)

                        decoded_messages.append(
                            Message(
                                coded_message=message,
                                service=service,
                                coding_object=gnr,
                                param_dict=decoded_gnr))
                    except DecodeError:
                        pass

                if not gnr_found:
                    raise e

        if len(decoded_messages) == 0:
            raise DecodeError(
                f"None of the services {[x.short_name for x in candidate_services]} could parse {message.hex()}."
            )

        return decoded_messages

    def decode(self, message: bytes) -> list[Message]:
        candidate_services = self._find_services_for_uds(message)

        return self._decode(message, candidate_services)

    def decode_response(self, response: bytes, request: bytes) -> list[Message]:
        candidate_services = self._find_services_for_uds(request)
        if candidate_services is None:
            raise DecodeError(f"Couldn't find corresponding service for request {request.hex()}.")

        return self._decode(response, candidate_services)

    #####
    # </PDU decoding>
    #####



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from xml.etree import ElementTree

from ..odxtypes import DataType
from ..utils import dataclass_fields_asdict
from .compuconst import CompuConst
from .compuinversevalue import CompuInverseValue


@dataclass(kw_only=True)
class CompuDefaultValue(CompuConst):
    compu_inverse_value: CompuInverseValue | None = None

    @staticmethod
    def compuvalue_from_et(et_element: ElementTree.Element, *,
                           data_type: DataType) -> "CompuDefaultValue":
        kwargs = dataclass_fields_asdict(
            CompuConst.compuvalue_from_et(et_element, data_type=data_type))

        compu_inverse_value = None
        if (civ_elem := et_element.find("COMPU-INVERSE-VALUE")) is not None:
            compu_inverse_value = CompuInverseValue.compuvalue_from_et(
                civ_elem, data_type=data_type)

        return CompuDefaultValue(**kwargs, compu_inverse_value=compu_inverse_value)



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import cast
from xml.etree import ElementTree

from ..exceptions import DecodeError, EncodeError, odxassert, odxraise
from ..odxdoccontext import OdxDocContext
from ..odxtypes import AtomicOdxType, DataType
from ..progcode import ProgCode
from ..utils import dataclass_fields_asdict
from .compucategory import CompuCategory
from .compumethod import CompuMethod


@dataclass(kw_only=True)
class CompuCodeCompuMethod(CompuMethod):
    """A compu method specifies the tranfer functions using Java bytecode

    For details, refer to ASAM specification MCD-2 D (ODX), section 7.3.6.6.9.
    """

    @property
    def internal_to_phys_code(self) -> ProgCode | None:
        if self.compu_internal_to_phys is None:
            return None

        return self.compu_internal_to_phys.prog_code

    @property
    def phys_to_internal_code(self) -> ProgCode | None:
        if self.compu_phys_to_internal is None:
            return None

        return self.compu_phys_to_internal.prog_code

    @staticmethod
    def compu_method_from_et(et_element: ElementTree.Element, context: OdxDocContext, *,
                             internal_type: DataType,
                             physical_type: DataType) -> "CompuCodeCompuMethod":
        cm = CompuMethod.compu_method_from_et(
            et_element, context, internal_type=internal_type, physical_type=physical_type)
        kwargs = dataclass_fields_asdict(cm)

        return CompuCodeCompuMethod(**kwargs)

    def __post_init__(self) -> None:
        odxassert(self.category == CompuCategory.COMPUCODE,
                  "CompuCodeCompuMethod must exhibit COMPUCODE category")

    def convert_internal_to_physical(self, internal_value: AtomicOdxType) -> AtomicOdxType:
        odxraise(r"CompuCodeCompuMethod cannot be executed by odxtools", DecodeError)
        return cast(AtomicOdxType, None)

    def convert_physical_to_internal(self, physical_value: AtomicOdxType) -> AtomicOdxType:
        odxraise(r"CompuCodeCompuMethod cannot be executed by odxtools", EncodeError)
        return cast(AtomicOdxType, None)

    def is_valid_physical_value(self, physical_value: AtomicOdxType) -> bool:
        # CompuCodeCompuMethod cannot be executed by odxtools
        return False

    def is_valid_internal_value(self, internal_value: AtomicOdxType) -> bool:
        # CompuCodeCompuMethod cannot be executed by odxtools
        return False



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import Any, cast
from xml.etree import ElementTree

from typing_extensions import override

from ..dataobjectproperty import DataObjectProperty
from ..decodestate import DecodeState
from ..dopbase import DopBase
from ..dtcdop import DtcDop
from ..encodestate import EncodeState
from ..exceptions import odxassert, odxrequire
from ..odxdoccontext import OdxDocContext
from ..odxlink import OdxLinkDatabase, OdxLinkId, OdxLinkRef, resolve_snref
from ..odxtypes import AtomicOdxType, ParameterValue
from ..physicaltype import PhysicalType
from ..snrefcontext import SnRefContext
from ..utils import dataclass_fields_asdict
from .parameter import Parameter


@dataclass(kw_only=True)
class ParameterWithDOP(Parameter):
    dop_ref: OdxLinkRef | None = None
    dop_snref: str | None = None

    @property
    def dop(self) -> DopBase:
        """This is usually a DataObjectProperty or a Structure object"""

        return self._dop

    @staticmethod
    @override
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "ParameterWithDOP":

        kwargs = dataclass_fields_asdict(Parameter.from_et(et_element, context))

        dop_ref = OdxLinkRef.from_et(et_element.find("DOP-REF"), context)
        dop_snref = None
        if (dop_snref_elem := et_element.find("DOP-SNREF")) is not None:
            dop_snref = odxrequire(dop_snref_elem.get("SHORT-NAME"))

        return ParameterWithDOP(dop_ref=dop_ref, dop_snref=dop_snref, **kwargs)

    def __post_init__(self) -> None:
        odxassert(self.dop_snref is not None or self.dop_ref is not None,
                  f"Param {self.short_name} without a DOP-(SN)REF should not exist!")
        self._dop: DopBase

    @override
    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        return super()._build_odxlinks()

    @override
    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        super()._resolve_odxlinks(odxlinks)

        if self.dop_ref is not None:
            odxassert(self.dop_snref is None)
            self._dop = odxlinks.resolve(self.dop_ref)

    @override
    def _resolve_snrefs(self, context: SnRefContext) -> None:
        super()._resolve_snrefs(context)

        if self.dop_snref:
            ddds = odxrequire(context.diag_layer).diag_data_dictionary_spec
            self._dop = resolve_snref(self.dop_snref, ddds.all_data_object_properties, DopBase)

    @override
    def get_static_bit_length(self) -> int | None:
        if self._dop is not None:
            return self._dop.get_static_bit_length()
        else:
            return None

    @property
    def physical_type(self) -> PhysicalType | None:
        if isinstance(self.dop, (DataObjectProperty, DtcDop)):
            return self.dop.physical_type
        else:
            return None

    @override
    def _encode_positioned_into_pdu(self, physical_value: ParameterValue | None,
                                    encode_state: EncodeState) -> None:
        self.dop.encode_into_pdu(cast(AtomicOdxType, physical_value), encode_state)

    @override
    def _decode_positioned_from_pdu(self, decode_state: DecodeState) -> ParameterValue:
        return self.dop.decode_from_pdu(decode_state)



# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from typing import Any
from xml.etree import ElementTree

from .dtcconnector import DtcConnector
from .element import IdentifiableElement
from .envdataconnector import EnvDataConnector
from .nameditemlist import NamedItemList
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId
from .snrefcontext import SnRefContext
from .subcomponentparamconnector import SubComponentParamConnector
from .subcomponentpattern import SubComponentPattern
from .tablerowconnector import TableRowConnector
from .utils import dataclass_fields_asdict


@dataclass(kw_only=True)
class SubComponent(IdentifiableElement):
    """Sub-components describe collections of related diagnostic variables

    Note that the communication paradigm via diagnostic variables is
    somewhat uncommon. If your ECU does not define any, there's no
    need for it to define sub-components.

    """

    sub_component_patterns: list[SubComponentPattern] = field(default_factory=list)
    sub_component_param_connectors: NamedItemList[SubComponentParamConnector] = field(
        default_factory=NamedItemList)
    table_row_connectors: NamedItemList[TableRowConnector] = field(default_factory=NamedItemList)
    env_data_connectors: NamedItemList[EnvDataConnector] = field(default_factory=NamedItemList)
    dtc_connectors: NamedItemList[DtcConnector] = field(default_factory=NamedItemList)

    semantic: str | None = None

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "SubComponent":
        kwargs = dataclass_fields_asdict(IdentifiableElement.from_et(et_element, context))

        semantic = et_element.get("SEMANTIC")

        sub_component_patterns = [
            SubComponentPattern.from_et(el, context)
            for el in et_element.iterfind("SUB-COMPONENT-PATTERNS/SUB-COMPONENT-PATTERN")
        ]
        sub_component_param_connectors = [
            SubComponentParamConnector.from_et(el, context) for el in et_element.iterfind(
                "SUB-COMPONENT-PARAM-CONNECTORS/SUB-COMPONENT-PARAM-CONNECTOR")
        ]
        table_row_connectors = [
            TableRowConnector.from_et(el, context)
            for el in et_element.iterfind("TABLE-ROW-CONNECTORS/TABLE-ROW-CONNECTOR")
        ]
        env_data_connectors = [
            EnvDataConnector.from_et(el, context)
            for el in et_element.iterfind("ENV-DATA-CONNECTORS/ENV-DATA-CONNECTOR")
        ]
        dtc_connectors = [
            DtcConnector.from_et(el, context)
            for el in et_element.iterfind("DTC-CONNECTORS/DTC-CONNECTOR")
        ]

        return SubComponent(
            semantic=semantic,
            sub_component_patterns=sub_component_patterns,
            sub_component_param_connectors=NamedItemList(sub_component_param_connectors),
            table_row_connectors=NamedItemList(table_row_connectors),
            env_data_connectors=NamedItemList(env_data_connectors),
            dtc_connectors=NamedItemList(dtc_connectors),
            **kwargs)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        result = {}

        for scp in self.sub_component_patterns:
            result.update(scp._build_odxlinks())

        for scpc in self.sub_component_param_connectors:
            result.update(scpc._build_odxlinks())

        for trc in self.table_row_connectors:
            result.update(trc._build_odxlinks())

        for edc in self.env_data_connectors:
            result.update(edc._build_odxlinks())

        for dtc_conn in self.dtc_connectors:
            result.update(dtc_conn._build_odxlinks())

        return result

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        for scp in self.sub_component_patterns:
            scp._resolve_odxlinks(odxlinks)

        for scpc in self.sub_component_param_connectors:
            scpc._resolve_odxlinks(odxlinks)

        for trc in self.table_row_connectors:
            trc._resolve_odxlinks(odxlinks)

        for edc in self.env_data_connectors:
            edc._resolve_odxlinks(odxlinks)

        for dtc_conn in self.dtc_connectors:
            dtc_conn._resolve_odxlinks(odxlinks)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        for scp in self.sub_component_patterns:
            scp._resolve_snrefs(context)

        for scpc in self.sub_component_param_connectors:
            scpc._resolve_snrefs(context)

        for trc in self.table_row_connectors:
            trc._resolve_snrefs(context)

        for edc in self.env_data_connectors:
            edc._resolve_snrefs(context)

        for dtc_conn in self.dtc_connectors:
            dtc_conn._resolve_snrefs(context)



# SPDX-License-Identifier: MIT
from enum import Enum


class TransMode(Enum):
    SEND_ONLY = "SEND-ONLY"
    RECEIVE_ONLY = "RECEIVE-ONLY"
    SEND_AND_RECEIVE = "SEND-AND-RECEIVE"
    SEND_OR_RECEIVE = "SEND-OR-RECEIVE"



# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from typing import Any
from xml.etree import ElementTree

from ..comparamspec import ComparamSpec
from ..exceptions import odxrequire
from ..odxdoccontext import OdxDocContext
from ..odxlink import OdxLinkDatabase, OdxLinkId, OdxLinkRef, resolve_snref
from ..parentref import ParentRef
from ..protstack import ProtStack
from ..snrefcontext import SnRefContext
from ..utils import dataclass_fields_asdict
from .hierarchyelementraw import HierarchyElementRaw


@dataclass(kw_only=True)
class ProtocolRaw(HierarchyElementRaw):
    """This is the base class for diagnostic layers that describe a
    protocol which can be used to communicate with an ECU

    This class represents the data present in the XML, not the "logical" view.

    """

    comparam_spec_ref: OdxLinkRef
    prot_stack_snref: str | None = None
    parent_refs: list[ParentRef] = field(default_factory=list)

    @property
    def comparam_spec(self) -> ComparamSpec:
        return self._comparam_spec

    @property
    def prot_stack(self) -> ProtStack | None:
        return self._prot_stack

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "ProtocolRaw":
        # objects contained by diagnostic layers exibit an additional
        # document fragment for the diag layer, so we use the document
        # fragments of the odx id of the diag layer for IDs of
        # contained objects.
        her = HierarchyElementRaw.from_et(et_element, context)
        kwargs = dataclass_fields_asdict(her)

        comparam_spec_ref = OdxLinkRef.from_et(
            odxrequire(et_element.find("COMPARAM-SPEC-REF")), context)

        prot_stack_snref = None
        if (prot_stack_snref_elem := et_element.find("PROT-STACK-SNREF")) is not None:
            prot_stack_snref = odxrequire(prot_stack_snref_elem.get("SHORT-NAME"))

        parent_refs = [
            ParentRef.from_et(pr_el, context)
            for pr_el in et_element.iterfind("PARENT-REFS/PARENT-REF")
        ]

        return ProtocolRaw(
            comparam_spec_ref=comparam_spec_ref,
            prot_stack_snref=prot_stack_snref,
            parent_refs=parent_refs,
            **kwargs)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        result = super()._build_odxlinks()

        for parent_ref in self.parent_refs:
            result.update(parent_ref._build_odxlinks())

        return result

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        super()._resolve_odxlinks(odxlinks)

        for parent_ref in self.parent_refs:
            parent_ref._resolve_odxlinks(odxlinks)

        self._comparam_spec = odxlinks.resolve(self.comparam_spec_ref, ComparamSpec)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        super()._resolve_snrefs(context)

        self._prot_stack = None
        if self.prot_stack_snref is not None:
            self._prot_stack = resolve_snref(self.prot_stack_snref, self._comparam_spec.prot_stacks,
                                             ProtStack)

        for parent_ref in self.parent_refs:
            parent_ref._resolve_snrefs(context)



# SPDX-License-Identifier: MIT
from enum import Enum


class StandardizationLevel(Enum):
    STANDARD = "STANDARD"
    OEM_SPECIFIC = "OEM-SPECIFIC"
    OPTIONAL = "OPTIONAL"
    OEM_OPTIONAL = "OEM-OPTIONAL"



from typing import TYPE_CHECKING, NamedTuple

from packaging.version import Version

if TYPE_CHECKING:
    from odxtools.odxlink import OdxDocFragment


class OdxDocContext(NamedTuple):
    version: Version
    doc_fragments: list["OdxDocFragment"]



# SPDX-License-Identifier: MIT
import warnings
from dataclasses import dataclass
from typing import Any, cast
from xml.etree import ElementTree

from typing_extensions import override

from ..createanydiagcodedtype import create_any_diag_coded_type_from_et
from ..decodestate import DecodeState
from ..diagcodedtype import DiagCodedType
from ..encodestate import EncodeState
from ..exceptions import DecodeError, EncodeError, odxraise, odxrequire
from ..odxdoccontext import OdxDocContext
from ..odxlink import OdxLinkId
from ..odxtypes import AtomicOdxType, DataType, ParameterValue
from ..utils import dataclass_fields_asdict
from .parameter import Parameter, ParameterType


@dataclass(kw_only=True)
class CodedConstParameter(Parameter):

    coded_value_raw: str
    diag_coded_type: DiagCodedType

    @property
    @override
    def parameter_type(self) -> ParameterType:
        return "CODED-CONST"

    @property
    @override
    def is_required(self) -> bool:
        return False

    @property
    @override
    def is_settable(self) -> bool:
        return False

    @property
    def coded_value(self) -> AtomicOdxType:
        return self._coded_value

    @staticmethod
    @override
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "CodedConstParameter":

        kwargs = dataclass_fields_asdict(Parameter.from_et(et_element, context))

        coded_value_raw = odxrequire(et_element.findtext("CODED-VALUE"))
        dct_elem = odxrequire(et_element.find("DIAG-CODED-TYPE"))
        diag_coded_type = create_any_diag_coded_type_from_et(dct_elem, context)

        return CodedConstParameter(
            coded_value_raw=coded_value_raw, diag_coded_type=diag_coded_type, **kwargs)

    def __post_init__(self) -> None:
        self._coded_value = cast(
            AtomicOdxType, self.diag_coded_type.base_data_type.from_string(self.coded_value_raw))

    @override
    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        result = super()._build_odxlinks()

        result.update(self.diag_coded_type._build_odxlinks())

        return result

    @override
    def get_static_bit_length(self) -> int | None:
        return self.diag_coded_type.get_static_bit_length()

    @property
    def internal_data_type(self) -> DataType:
        return self.diag_coded_type.base_data_type

    @override
    def _encode_positioned_into_pdu(self, physical_value: ParameterValue | None,
                                    encode_state: EncodeState) -> None:
        if physical_value is not None and physical_value != self.coded_value:
            odxraise(
                f"Value for constant parameter `{self.short_name}` name can "
                f"only be specified as {self.coded_value!r} (is: {physical_value!r})", EncodeError)

        internal_value = self.coded_value

        self.diag_coded_type.encode_into_pdu(
            internal_value=internal_value, encode_state=encode_state)

    @override
    def _decode_positioned_from_pdu(self, decode_state: DecodeState) -> AtomicOdxType:
        coded_val = self.diag_coded_type.decode_from_pdu(decode_state)

        # Check if the coded value contained by the message is correct.
        if self.coded_value != coded_val:
            warnings.warn(
                f"Coded constant parameter does not match! "
                f"The parameter {self.short_name} expected coded "
                f"value {str(self.coded_value)} but got {str(coded_val)} "
                f"at byte position {decode_state.cursor_byte_position} "
                f"in coded message {decode_state.coded_message.hex()}.",
                DecodeError,
                stacklevel=1,
            )

        return coded_val

    def get_description_of_valid_values(self) -> str:
        """return a human-understandable description of valid physical values"""
        return f"Constant internal value: {str(self.coded_value)}"



#!/usr/bin/env python
# SPDX-License-Identifier: MIT

import argparse
import os
from typing import Any, cast

from rich import print as rich_print
from rich.padding import Padding as RichPadding
from rich.table import Table as RichTable

from ..database import Database
from ..diaglayers.diaglayer import DiagLayer
from ..diagservice import DiagService
from ..loadfile import load_file
from ..odxtypes import AtomicOdxType
from ..parameters.codedconstparameter import CodedConstParameter
from ..parameters.nrcconstparameter import NrcConstParameter
from ..parameters.parameter import Parameter
from ..parameters.physicalconstantparameter import PhysicalConstantParameter
from ..parameters.valueparameter import ValueParameter
from . import _parser_utils
from ._parser_utils import SubparsersList
from ._print_utils import (extract_service_tabulation_data, print_dl_metrics,
                           print_service_parameters)

# name of the tool
_odxtools_tool_name_ = "compare"

VariantName = str
VariantType = str
NewServices = list[DiagService]
DeletedServices = list[DiagService]
RenamedServices = list[list[str | DiagService]]
ServicesWithParamChanges = list[list[str | DiagService]]

SpecsServiceDict = dict[str, VariantName | VariantType | NewServices | DeletedServices
                        | RenamedServices | ServicesWithParamChanges]

NewVariants = list[DiagLayer]
DeletedVariants = list[DiagLayer]

SpecsChangesVariants = dict[str, NewVariants | DeletedVariants | SpecsServiceDict]


class Display:
    # class with variables and functions to display the result of the comparison

    # TODO
    # - Idea: results as json export
    #  - write results of comparison in json structure
    #  - use odxlinks to refer to dignostic services / objects if
    #    changes have already been detected (e.g. in another ecu
    #    variant / diagnostic layer)
    # - print all information about parameter properties (request,
    #   pos. response & neg. response parameters) for changed diagnostic
    #   services
    param_detailed: bool
    obj_detailed: bool

    def __init__(self) -> None:
        pass

    def print_dl_changes(self, service_dict: SpecsServiceDict) -> None:

        if service_dict["new_services"] or service_dict["deleted_services"] or service_dict[
                "changed_name_of_service"][0] or service_dict["changed_parameters_of_service"][0]:
            assert isinstance(service_dict["diag_layer"], str)
            rich_print()
            rich_print(
                f"Changed diagnostic services for diagnostic layer '{service_dict['diag_layer']}' ({service_dict['diag_layer_type']}):"
            )
        if service_dict["new_services"]:
            assert isinstance(service_dict["new_services"], list)
            rich_print()
            rich_print(" [blue]New services[/blue]")
            rich_print(extract_service_tabulation_data(
                service_dict["new_services"]))  # type: ignore[arg-type]
        if service_dict["deleted_services"]:
            assert isinstance(service_dict["deleted_services"], list)
            rich_print()
            rich_print(" [blue]Deleted services[/blue]")
            rich_print(extract_service_tabulation_data(
                service_dict["deleted_services"]))  # type: ignore[arg-type]
        if service_dict["changed_name_of_service"][0]:
            rich_print()
            rich_print(" [blue]Renamed services[/blue]")
            rich_print(extract_service_tabulation_data(
                service_dict["changed_name_of_service"][0]))  # type: ignore[arg-type]
        if service_dict["changed_parameters_of_service"][0]:
            rich_print()
            rich_print(" [blue]Services with parameter changes[/blue]")
            # create table with information about services with parameter changes
            changed_param_column = [
                str(x) for x in service_dict["changed_parameters_of_service"][
                    1]  # type: ignore[union-attr]
            ]
            table = extract_service_tabulation_data(
                service_dict["changed_parameters_of_service"][0],  # type: ignore[arg-type]
                additional_columns=[("Changed Parameters", changed_param_column)])
            rich_print(table)

            for service_idx, service in enumerate(
                    service_dict["changed_parameters_of_service"][0]):  # type: ignore[arg-type]
                assert isinstance(service, DiagService)
                rich_print()
                rich_print(
                    f"  Detailed changes of diagnostic service [u cyan]{service.short_name}[/u cyan]"
                )
                #   detailed_info in [infotext1, dict1, infotext2, dict2, ...]
                info_list = cast(
                    list,  # type: ignore[type-arg]
                    service_dict["changed_parameters_of_service"])[2][service_idx]
                for detailed_info in info_list:
                    if isinstance(detailed_info, str):
                        rich_print()
                        rich_print(detailed_info)
                    elif isinstance(detailed_info, dict):
                        table = RichTable(
                            show_header=True,
                            header_style="bold cyan",
                            border_style="blue",
                            show_lines=True)
                        for header in detailed_info:
                            table.add_column(header)
                        rows = zip(*detailed_info.values(), strict=False)
                        for row in rows:
                            table.add_row(*map(str, row))

                        rich_print(RichPadding(table, pad=(0, 0, 0, 4)))
                        rich_print()
                if self.param_detailed:
                    # print all parameter details of diagnostic service
                    print_service_parameters(service, allow_unknown_bit_lengths=True)

    def print_database_changes(self, changes_variants: SpecsChangesVariants) -> None:
        # prints result of database comparison (input variable: dictionary: changes_variants)

        # diagnostic layers
        if changes_variants["new_diagnostic_layers"] or changes_variants[
                "deleted_diagnostic_layers"]:
            rich_print()
            rich_print("[bright_blue]Changed diagnostic layers[/bright_blue]: ")
            rich_print(" New diagnostic layers: ")
            for variant in changes_variants["new_diagnostic_layers"]:
                assert isinstance(variant, DiagLayer)
                rich_print(
                    f"  [magenta]{variant.short_name}[/magenta] ({variant.variant_type.value})")
            rich_print(" Deleted diagnostic layers: ")
            for variant in changes_variants["deleted_diagnostic_layers"]:
                assert isinstance(variant, DiagLayer)
                rich_print(
                    f"  [magenta]{variant.short_name}[/magenta] ({variant.variant_type.value})")

        # diagnostic services
        for _, value in changes_variants.items():
            if isinstance(value, dict):
                self.print_dl_changes(value)


class Comparison(Display):
    databases: list[Database] = []  # storage of database objects
    diagnostic_layers: list[DiagLayer] = []  # storage of DiagLayer objects
    diagnostic_layer_names: set[str] = set()  # storage of diagnostic layer names

    db_indicator_1: int
    db_indicator_2: int

    def __init__(self) -> None:
        pass

    def compare_parameters(self, param1: Parameter, param2: Parameter) -> dict[str, Any]:
        # checks whether properties of param1 and param2 differ
        # checked properties: Name, Byte Position, Bit Length, Semantic, Parameter Type, Value (Coded, Constant, Default etc.), Data Type, Data Object Property (Name, Physical Data Type, Unit)

        property = []
        old = []
        new = []

        def append_list(property_name: str, new_property_value: AtomicOdxType | None,
                        old_property_value: AtomicOdxType | None) -> None:
            property.append(property_name)
            old.append(old_property_value)
            new.append(new_property_value)

        if param1.short_name != param2.short_name:
            append_list("Parameter name", param1.short_name, param2.short_name)
        if param1.byte_position != param2.byte_position:
            append_list("Byte position", param1.byte_position, param2.byte_position)
        if param1.get_static_bit_length() != param2.get_static_bit_length():
            append_list("Bit Length", param1.get_static_bit_length(),
                        param2.get_static_bit_length())
        if param1.semantic != param2.semantic:
            append_list("Semantic", param1.semantic, param2.semantic)
        if param1.parameter_type != param2.parameter_type:
            append_list("Parameter type", param1.parameter_type, param2.parameter_type)

        if isinstance(param1, CodedConstParameter) and isinstance(param2, CodedConstParameter):
            if param1.diag_coded_type.base_data_type != param2.diag_coded_type.base_data_type:
                append_list("Data type", param1.diag_coded_type.base_data_type.name,
                            param2.diag_coded_type.base_data_type.name)
            if param1.coded_value != param2.coded_value:
                if isinstance(param1.coded_value, int) and isinstance(param2.coded_value, int):
                    append_list(
                        "Value",
                        f"0x{param1.coded_value:0{(param1.get_static_bit_length() or 0) // 4}X}",
                        f"0x{param2.coded_value:0{(param2.get_static_bit_length() or 0) // 4}X}")
                else:
                    append_list("Value", f"{param1.coded_value!r}", f"{param2.coded_value!r}")

        elif isinstance(param1, NrcConstParameter) and isinstance(param2, NrcConstParameter):
            if param1.diag_coded_type.base_data_type != param2.diag_coded_type.base_data_type:
                append_list("Data type", param1.diag_coded_type.base_data_type.name,
                            param2.diag_coded_type.base_data_type.name)
            if param1.coded_values != param2.coded_values:
                append_list("Values", str(param1.coded_values), str(param2.coded_values))

        elif (dop_1 := getattr(param1, "dop", None)) is not None and (dop_2 := getattr(
                param2, "dop", None)) is not None:

            if dop_1 != dop_2:
                # TODO: compare INTERNAL-CONSTR, COMPU-INTERNAL-TO-PHYS of DOP
                append_list("Linked DOP object", "", "")

                # DOP Name
                if dop_1.short_name != dop_2.short_name:
                    append_list(" DOP name", dop_1.short_name, dop_2.short_name)

                # DOP Unit
                if getattr(dop_1, "unit", None) and getattr(dop_2, "unit", None):
                    # (properties of unit object: short_name, long_name, description, odx_id, display_name, oid, factor_si_to_unit, offset_si_to_unit, physical_dimension_ref)
                    if dop_1.unit != dop_2.unit and dop_1.unit.short_name != dop_2.unit.short_name:
                        append_list("  DOP unit name", dop_1.unit.short_name, dop_2.unit.short_name)
                    elif dop_1.unit != dop_2.unit and dop_1.unit.display_name != dop_2.unit.display_name:
                        append_list("  DOP unit display name", dop_1.unit.display_name,
                                    dop_2.unit.display_name)
                    elif dop_1.unit != dop_2.unit:
                        append_list(" DOP unit object", "", "")

                if hasattr(dop_1, "physical_type") and hasattr(dop_2, "physical_type"):
                    if (dop_1.physical_type and dop_2.physical_type and
                            dop_1.physical_type.base_data_type
                            != dop_2.physical_type.base_data_type):
                        append_list(" DOP physical data type",
                                    dop_1.physical_type.base_data_type.name,
                                    dop_2.physical_type.base_data_type.name)

            if (isinstance(param1, PhysicalConstantParameter) and
                    isinstance(param2, PhysicalConstantParameter) and
                    param1.physical_constant_value != param2.physical_constant_value):
                if isinstance(param1.physical_constant_value, int) and isinstance(
                        param2.physical_constant_value, int):
                    append_list(
                        "Constant value",
                        f"0x{param1.physical_constant_value:0{(param1.get_static_bit_length() or 0) // 4}X}",
                        f"0x{param2.physical_constant_value:0{(param2.get_static_bit_length() or 0) // 4}X}"
                    )
                else:
                    append_list("Constant value", f"{param1.physical_constant_value!r}",
                                f"{param2.physical_constant_value!r}")

            elif (isinstance(param1, ValueParameter) and isinstance(param2, ValueParameter) and
                  param1.physical_default_value is not None and
                  param2.physical_default_value is not None and
                  param1.physical_default_value != param2.physical_default_value):
                if isinstance(param1.physical_default_value, int) and isinstance(
                        param2.physical_default_value, int):
                    append_list(
                        "Default value",
                        f"0x{param1.physical_default_value:0{(param1.get_static_bit_length() or 0) // 4}X}",
                        f"0x{param2.physical_default_value:0{(param2.get_static_bit_length() or 0) // 4}X}"
                    )
                else:
                    append_list("Default value", f"{param1.physical_default_value!r}",
                                f"{param2.physical_default_value!r}")

        return {"Property": property, "Old Value": old, "New Value": new}

    def compare_services(self, service1: DiagService,
                         service2: DiagService) -> list[SpecsServiceDict]:
        # compares request, positive response and negative response parameters of two diagnostic services

        information: list[str | dict[str, Any]] = [
        ]  # information = [infotext1, table1, infotext2, table2, ...]
        changed_params = ""

        # Request
        if service1.request is not None and service2.request is not None and len(
                service1.request.parameters) == len(service2.request.parameters):
            for res1_idx, param1 in enumerate(service1.request.parameters):
                for res2_idx, param2 in enumerate(service2.request.parameters):
                    if res1_idx == res2_idx:
                        # find changed request parameter properties
                        table = self.compare_parameters(param1, param2)
                        infotext = (f"   Properties of request parameter '{param2.short_name}' "
                                    f"that have changed:\n")
                        # array index starts with 0 -> param[0] is 1. service parameter

                        if table["Property"]:
                            information.append(infotext)
                            information.append(table)
                            changed_params += f"request parameter '{param2.short_name}',\n"
        else:
            changed_params += "request parameter list, "
            # infotext
            information.append(f"List of request parameters for service '{service2.short_name}' "
                               f"is not identical.\n")

            # table

            param_list1 = [] if service1.request is None else service1.request.parameters
            param_list2 = [] if service2.request is None else service2.request.parameters

            information.append({
                "List": ["Old list", "New list"],
                "Values": [f"\\{param_list1}", f"\\{param_list2}"]
            })

        # Positive Responses
        if len(service1.positive_responses) == len(service2.positive_responses):
            for res1_idx, response1 in enumerate(service1.positive_responses):
                for res2_idx, response2 in enumerate(service2.positive_responses):
                    if res1_idx == res2_idx:
                        if len(response1.parameters) == len(response2.parameters):
                            for param1_idx, param1 in enumerate(response1.parameters):
                                for param2_idx, param2 in enumerate(response2.parameters):
                                    if param1_idx == param2_idx:
                                        # find changed positive response parameter properties
                                        table = self.compare_parameters(param1, param2)
                                        infotext = (
                                            f"   Properties of positive response parameter '{param2.short_name}' that "
                                            f"have changed:\n")
                                        # array index starts with 0 -> param[0] is first service parameter

                                        if table["Property"]:
                                            information.append(infotext)
                                            information.append(table)
                                            changed_params += f"positive response parameter '{param2.short_name}',\n"
                        else:
                            changed_params += "positive response parameter list, "
                            # infotext
                            information.append(
                                f"List of positive response parameters for service '{service2.short_name}' is not identical."
                            )
                            # table
                            information.append({
                                "List": ["Old list", "New list"],
                                "Values": [str(response1.parameters),
                                           str(response2.parameters)]
                            })
        else:
            changed_params += "positive responses list, "
            # infotext
            information.append(
                f"List of positive responses for service '{service2.short_name}' is not identical.")
            # table
            information.append({
                "List": ["Old list", "New list"],
                "Values": [str(service1.positive_responses),
                           str(service2.positive_responses)]
            })

        # Negative Responses
        if len(service1.negative_responses) == len(service2.negative_responses):
            for res1_idx, response1 in enumerate(service1.negative_responses):
                for res2_idx, response2 in enumerate(service2.negative_responses):
                    if res1_idx == res2_idx:
                        if len(response1.parameters) == len(response2.parameters):
                            for param1_idx, param1 in enumerate(response1.parameters):
                                for param2_idx, param2 in enumerate(response2.parameters):
                                    if param1_idx == param2_idx:
                                        # find changed negative response parameter properties
                                        table = self.compare_parameters(param1, param2)
                                        infotext = f"   Properties of response parameter '{param2.short_name}' that have changed:\n"
                                        # array index starts with 0 -> param[0] is 1. service parameter

                                        if table["Property"]:
                                            information.append(infotext)
                                            information.append(table)
                                            changed_params += f"negative response parameter '{param2.short_name}',\n"
                        else:
                            changed_params += "positive response parameter list, "
                            # infotext
                            information.append(
                                f"List of positive response parameters for service '{service2.short_name}' is not identical.\n"
                            )
                            # table
                            information.append({
                                "List": ["Old list", "New list"],
                                "Values": [str(response1.parameters),
                                           str(response2.parameters)]
                            })
        else:
            changed_params += "negative responses list, "
            # infotext
            information.append(
                f"List of positive responses for service '{service2.short_name}' is not identical.\n"
            )
            # table
            information.append({
                "List": ["Old list", "New list"],
                "Values": [str(service1.negative_responses),
                           str(service2.negative_responses)]
            })

        return [information, changed_params]  # type: ignore[list-item]

    def compare_diagnostic_layers(self, dl1: DiagLayer,
                                  dl2: DiagLayer) -> dict:  # type: ignore[type-arg]
        # compares diagnostic services of two diagnostic layers with each other
        # save changes in dictionary (service_dict)
        # TODO: add comparison of SingleECUJobs

        new_services: NewServices = []
        deleted_services: DeletedServices = []
        renamed_service: RenamedServices = [[],
                                            []]  # TODO: implement list of (str, DiagService)-tuples
        services_with_param_changes: ServicesWithParamChanges = [
            [], [], []
        ]  # TODO: implement list of tuples (str, str, DiagService)-tuples

        service_dict: SpecsServiceDict = {
            "diag_layer": dl1.short_name,
            "diag_layer_type": dl1.variant_type.value,
            # list with added diagnostic services [service1, service2, service3, ...] Type: DiagService
            "new_services": new_services,
            # list with deleted diagnostic services [service1, service2, service3, ...] Type: DiagService
            "deleted_services": deleted_services,
            # list with diagnostic services where the service name changed [[services], [old service names]]
            "changed_name_of_service": renamed_service,
            # list with diagnostic services where the service parameter changed [[services], [changed_parameters], [information_texts]]
            "changed_parameters_of_service": services_with_param_changes
        }
        # service_dict["changed_name_of_service"][{0 = services, 1 = old service names}][i]
        # service_dict["changed_parameters_of_service"][{0 = services, 1 = changed_parameters, 2 = information_texts}][i]

        dl1_service_names = [service.short_name for service in dl1.services]

        # extract the constant prefixes for the requests of all
        # services (used for duck-typed rename detection)
        dl1_request_prefixes: list[bytes | None] = [
            None if s.request is None else s.request.coded_const_prefix() for s in dl1.services
        ]
        dl2_request_prefixes: list[bytes | None] = [
            None if s.request is None else s.request.coded_const_prefix() for s in dl2.services
        ]

        # compare diagnostic services
        for service1 in dl1.services:

            # check for added diagnostic services
            rq_prefix: bytes | None = None
            if service1.request is not None:
                rq_prefix = service1.request.coded_const_prefix()

            if service1 not in dl2.services:
                if rq_prefix is None or rq_prefix not in dl2_request_prefixes:
                    # TODO: this will not work in cases where the constant
                    # prefix of a request was modified...
                    service_dict["new_services"].append(  # type: ignore[union-attr]
                        service1)  # type: ignore[arg-type]

            # check whether names of diagnostic services have changed
            elif service1 not in dl2.services:
                if rq_prefix is None or rq_prefix in dl2_request_prefixes:
                    # get related diagnostic service for request
                    service2_idx = dl2_request_prefixes.index(rq_prefix)
                    service2 = dl2.services[service2_idx]

                    # save information about changes in dictionary

                    # add new service (type: DiagService)
                    service_dict["changed_name_of_service"][0].append(  # type: ignore[union-attr]
                        service1)
                    # add old service name (type: String)
                    service_dict["changed_name_of_service"][1].append(  # type: ignore[union-attr]
                        service2.short_name)

                    # compare request, pos. response and neg. response parameters of diagnostic services
                    detailed_information = self.compare_services(service1, service2)
                    # detailed_information = [[infotext1, table1, infotext2, table2, ...], changed_params]

                    # add information about changed diagnostic service parameters to dicitionary
                    if detailed_information[1]:  # check whether string "changed_params" is empty
                        # new service (type: DiagService)
                        service_dict["changed_parameters_of_service"][
                            0].append(  # type: ignore[union-attr]
                                service1)
                        # add parameters which have been changed (type: String)
                        service_dict["changed_parameters_of_service"][
                            1].append(  # type: ignore[union-attr]
                                detailed_information[1])  # type: ignore[arg-type]
                        # add detailed information about changed service parameters (type: list) [infotext1, table1, infotext2, table2, ...]
                        service_dict["changed_parameters_of_service"][
                            2].append(  # type: ignore[union-attr]
                                detailed_information[0])  # type: ignore[arg-type]

            for service2_idx, service2 in enumerate(dl2.services):

                # check for deleted diagnostic services
                if service2.short_name not in dl1_service_names and dl2_request_prefixes[
                        service2_idx] not in dl1_request_prefixes:

                    deleted_list = service_dict["deleted_services"]
                    assert isinstance(deleted_list, list)
                    if service2 not in deleted_list:
                        service_dict["deleted_services"].append(  # type: ignore[union-attr]
                            service2)  # type: ignore[arg-type]

                if service1.short_name == service2.short_name:
                    # compare request, pos. response and neg. response parameters of both diagnostic services
                    detailed_information = self.compare_services(service1, service2)
                    # detailed_information = [[infotext1, table1, infotext2, table2, ...], changed_params]

                    # add information about changed diagnostic service parameters to dicitionary
                    if detailed_information[1]:  # check whether string "changed_params" is empty
                        # new service (type: DiagService)
                        service_dict["changed_parameters_of_service"][
                            0].append(  # type: ignore[union-attr]
                                service1)
                        # add parameters which have been changed (type: String)
                        service_dict["changed_parameters_of_service"][  # type: ignore[union-attr]
                            1].append(detailed_information[1])  # type: ignore[arg-type]
                        # add detailed information about changed service parameters (type: list) [infotext1, table1, infotext2, table2, ...]
                        service_dict["changed_parameters_of_service"][  # type: ignore[union-attr]
                            2].append(detailed_information[0])  # type: ignore[arg-type]
        return service_dict

    def compare_databases(self, database_new: Database,
                          database_old: Database) -> dict:  # type: ignore[type-arg]
        # compares two PDX-files with each other

        new_variants: NewVariants = []
        deleted_variants: DeletedVariants = []

        changes_variants: SpecsChangesVariants = {
            "new_diagnostic_layers": new_variants,
            "deleted_diagnostic_layers": deleted_variants
        }

        # compare databases
        for _, dl1 in enumerate(database_new.diag_layers):
            # check for new diagnostic layers
            if dl1.short_name not in [dl.short_name for dl in database_old.diag_layers]:
                changes_variants["new_diagnostic_layers"].append(dl1)  # type: ignore[union-attr]

            for _, dl2 in enumerate(database_old.diag_layers):
                # check for deleted diagnostic layers
                if (dl2.short_name not in [dl.short_name for dl in database_new.diag_layers] and
                        dl2 not in changes_variants["deleted_diagnostic_layers"]):

                    changes_variants[
                        "deleted_diagnostic_layers"].append(  # type: ignore[union-attr]
                            dl2)

                if dl1.short_name == dl2.short_name and dl1.short_name in self.diagnostic_layer_names:
                    # compare diagnostic services of both diagnostic layers
                    # save diagnostic service changes in dictionary (empty if no changes)
                    service_dict: SpecsServiceDict = self.compare_diagnostic_layers(dl1, dl2)
                    if service_dict:
                        # adds information about diagnostic service changes to return variable (changes_variants)
                        changes_variants.update({dl1.short_name: service_dict})

        return changes_variants


def add_subparser(subparsers: SubparsersList) -> None:
    parser = subparsers.add_parser(
        "compare",
        description="\n".join([
            "Compares two versions of diagnostic layers or databases with each other. Checks whether diagnostic services and its parameters have changed.",
            "",
            "Examples:",
            "  Comparison of two diagnostic layers:",
            "    odxtools compare ./path/to/database.pdx -v variant1 variant2",
            "  Comparison of two database versions:",
            "    odxtools compare ./path/to/database.pdx -db ./path/to/old-database.pdx",
            "  For more information use:",
            "    odxtools compare -h",
        ]),
        help="Compares two versions of diagnostic layers and/or databases with each other. Checks whether diagnostic services and its parameters have changed.",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    _parser_utils.add_pdx_argument(parser)

    parser.add_argument(
        "-v",
        "--variants",
        nargs="+",
        metavar="VARIANT",
        required=False,
        default=None,
        help="Compare specified (ecu) variants to each other.",
    )

    parser.add_argument(
        "-db",
        "--database",
        nargs="+",
        default=None,
        metavar="DATABASE",
        required=False,
        help="Compare specified database file(s) to database file of first input argument.",
    )

    parser.add_argument(
        "-nd",
        "--no-details",
        action="store_false",
        default=True,
        required=False,
        help="Don't show all service parameter details",
    )

    # TODO
    # Idea: provide folder with multiple pdx files as argument
    # -> load all pdx files in folder, sort them alphabetically, compare databases pairwaise
    # -> calculate metrics (number of added services, number of changed services, number of removed services, total number of services per ecu variant, ...)
    # -> display metrics graphically


def run(args: argparse.Namespace) -> None:

    task = Comparison()
    task.param_detailed = args.no_details

    db_names = [args.pdx_file if isinstance(args.pdx_file, str) else str(args.pdx_file[0])]

    if args.database and args.variants:
        # compare specified databases, consider only specified variants

        for name in args.database:
            db_names.append(name) if isinstance(name, str) else str(name[0])

        task.databases = [load_file(name) for name in db_names]
        diag_layer_names = {dl.short_name for db in task.databases for dl in db.diag_layers}

        task.diagnostic_layer_names = diag_layer_names.intersection(set(args.variants))

        for name in args.variants:
            if name not in task.diagnostic_layer_names:
                rich_print(f"The variant '{name}' could not be found!")
                return

        task.db_indicator_1 = 0

        for db_idx, _ in enumerate(task.databases):
            if db_idx + 1 >= len(task.databases):
                break
            task.db_indicator_2 = db_idx + 1

            rich_print()
            rich_print(f"Changes in file '{os.path.basename(db_names[0])}'")
            rich_print(f" (compared to '{os.path.basename(db_names[db_idx + 1])}')")

            rich_print()
            rich_print(f"Overview of diagnostic layers (for {os.path.basename(db_names[0])})")
            print_dl_metrics([
                variant for variant in task.databases[0].diag_layers
                if variant.short_name in task.diagnostic_layer_names
            ])

            rich_print()
            rich_print(
                f"Overview of diagnostic layers (for {os.path.basename(db_names[db_idx+1])})")
            print_dl_metrics([
                variant for variant in task.databases[db_idx + 1].diag_layers
                if variant.short_name in task.diagnostic_layer_names
            ])

            task.print_database_changes(
                task.compare_databases(task.databases[0], task.databases[db_idx + 1]))

    elif args.database:
        # compare specified databases, consider all variants

        for name in args.database:
            db_names.append(name)
        task.databases = [load_file(name) for name in db_names]

        # collect all diagnostic layers from all specified databases
        task.diagnostic_layer_names = {
            dl.short_name
            for db in task.databases
            for dl in db.diag_layers
        }
        task.db_indicator_1 = 0

        for db_idx, _ in enumerate(task.databases):
            if db_idx + 1 >= len(task.databases):
                break
            task.db_indicator_2 = db_idx + 1

            rich_print()
            rich_print(f"Changes in file '{os.path.basename(db_names[0])}")
            rich_print(f" (compared to '{os.path.basename(db_names[db_idx + 1])}')")

            rich_print()
            rich_print(f"Overview of diagnostic layers (for {os.path.basename(db_names[0])})")
            print_dl_metrics(list(task.databases[0].diag_layers))

            rich_print()
            rich_print(
                f"Overview of diagnostic layers (for {os.path.basename(db_names[db_idx+1])})")
            print_dl_metrics(list(task.databases[db_idx + 1].diag_layers))

            task.print_database_changes(
                task.compare_databases(task.databases[0], task.databases[db_idx + 1]))

    elif args.variants:
        # no databases specified -> comparison of diagnostic layers

        odxdb = _parser_utils.load_file(args)
        task.databases = [odxdb]

        diag_layer_names = {dl.short_name for db in task.databases for dl in db.diag_layers}

        task.diagnostic_layer_names = diag_layer_names.intersection(set(args.variants))
        task.diagnostic_layers = [
            dl for db in task.databases for dl in db.diag_layers
            if dl.short_name in task.diagnostic_layer_names
        ]

        for name in args.variants:
            if name not in task.diagnostic_layer_names:
                rich_print(f"The variant '{name}' could not be found!")
                return

        rich_print()
        rich_print(f"Overview of diagnostic layers: ")
        print_dl_metrics(task.diagnostic_layers)

        for db_idx, dl in enumerate(task.diagnostic_layers):
            if db_idx + 1 >= len(task.diagnostic_layers):
                break

            rich_print()
            rich_print(f"Changes in diagnostic layer '{dl.short_name}' ({dl.variant_type.value})")
            rich_print(
                f" (compared to '{task.diagnostic_layers[db_idx + 1].short_name}' ({task.diagnostic_layers[db_idx + 1].variant_type.value}))"
            )
            task.print_dl_changes(
                task.compare_diagnostic_layers(dl, task.diagnostic_layers[db_idx + 1]))

    else:
        # no databases & no variants specified
        rich_print("Please specify either a database or variant for a comparison")



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import Any, cast
from xml.etree import ElementTree

from .diaglayers.diaglayer import DiagLayer
from .diagnostictroublecode import DiagnosticTroubleCode
from .diagservice import DiagService
from .exceptions import odxraise, odxrequire
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId, resolve_snref
from .odxtypes import BytesTypes, ParameterValue, ParameterValueDict
from .snrefcontext import SnRefContext


@dataclass(kw_only=True)
class MatchingParameter:
    """According to ISO 22901, a MatchingParameter contains a string
    value identifying the active ECU or base variant. Moreover, it
    references a DIAG-COMM via snref and one of its positive
    response's OUT-PARAM-IF via snref or snpathref.

    Unlike other parameters defined in the `parameters` package, a
    MatchingParameter is not transferred over the network.

    """

    expected_value: str
    diag_comm_snref: str

    # be aware that the checker rules 23 and 110 contradict each other
    # here: the first specifies that the referenced parameter must be
    # present in *all* positive responses of the referenced service,
    # whilst the latter mandates it to be present least one positive
    # or negative response. What it probably actually wants to say is
    # that any response that can possibly be received shall exhibit
    # the referenced parameter.
    out_param_if_snref: str | None = None
    out_param_if_snpathref: str | None = None

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "MatchingParameter":

        expected_value = odxrequire(et_element.findtext("EXPECTED-VALUE"))
        diag_comm_snref = odxrequire(
            odxrequire(et_element.find("DIAG-COMM-SNREF")).get("SHORT-NAME"))

        out_param_if_snref = None
        out_param_if_snpathref = None
        if (out_param_snref_el := et_element.find("OUT-PARAM-IF-SNREF")) is not None:
            out_param_if_snref = odxrequire(out_param_snref_el.get("SHORT-NAME"))
        elif (out_param_snpathref_el := et_element.find("OUT-PARAM-IF-SNPATHREF")) is not None:
            out_param_if_snpathref = odxrequire(out_param_snpathref_el.get("SHORT-NAME-PATH"))
        else:
            odxraise("Output parameter must not be left unspecified")

        return MatchingParameter(
            expected_value=expected_value,
            diag_comm_snref=diag_comm_snref,
            out_param_if_snref=out_param_if_snref,
            out_param_if_snpathref=out_param_if_snpathref,
        )

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        return {}

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        pass

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        # note that we do not resolve the SNREF to the diag layer here
        # because it is a component which changes depending on the
        # execution context and for variant patterns the execution
        # context is highly volatile w.r.t. the diag layer in question
        pass

    def get_ident_service(self, diag_layer: DiagLayer) -> DiagService:
        return resolve_snref(self.diag_comm_snref, diag_layer.services, DiagService)

    def matches(self, param_dict: ParameterValueDict) -> bool:
        """
        Returns true iff the provided identification value matches this MatchingParameter's
        expected value.
        """

        if self.out_param_if_snref is not None:
            snpath_chunks = [self.out_param_if_snref]
        elif self.out_param_if_snpathref is not None:
            snpath_chunks = self.out_param_if_snpathref.split(".")
        else:
            odxraise("no out_param_if specified")
            return False

        return self.__matches(param_dict, snpath_chunks)

    def __matches(self, param_dict: ParameterValue, snpath_chunks: list[str]) -> bool:
        if len(snpath_chunks) == 0:
            parameter_value = param_dict
            if isinstance(parameter_value, dict):
                odxraise(f"Parameter must not be a structure")
                return False

            if isinstance(parameter_value, float):
                # allow a slight tolerance if the expected value is
                # floating point
                return abs(float(self.expected_value) - parameter_value) < 1e-8
            elif isinstance(parameter_value, BytesTypes):
                return bytes(parameter_value).hex().upper() == self.expected_value.upper()
            elif isinstance(parameter_value, DiagnosticTroubleCode):
                # TODO: what happens if non-numerical DTCs like
                # "U123456" are specified? Is specifying DTCs even
                # allowed in variant patterns?
                return hex(parameter_value.trouble_code).upper() == self.expected_value.upper()
            else:
                return self.expected_value == str(parameter_value)

        if not isinstance(param_dict, dict):
            odxraise(f"Parameter {snpath_chunks[0]} must be a structure")
            return False

        sub_value = param_dict.get(snpath_chunks[0])
        if sub_value is None:
            return False

        if isinstance(sub_value, tuple) and len(sub_value) == 2:
            # table struct parameter
            sub_value = sub_value[1]

        if isinstance(sub_value, list):
            # the spec mandates that if any item of a field matches,
            # the parameter as a whole matches
            for x in sub_value:
                if self.__matches(x, snpath_chunks[1:]):
                    return True

            return False

        return self.__matches(cast(ParameterValue, sub_value), snpath_chunks[1:])



# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from typing import Any
from xml.etree import ElementTree

from .diagclasstype import DiagClassType
from .diagcomm import DiagComm
from .exceptions import odxassert, odxraise, odxrequire
from .nameditemlist import NamedItemList
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId, OdxLinkRef, resolve_snref
from .snrefcontext import SnRefContext
from .table import Table


@dataclass(kw_only=True)
class DynIdDefModeInfo:
    def_mode: str

    clear_dyn_def_message_ref: OdxLinkRef | None = None
    clear_dyn_def_message_snref: str | None = None

    read_dyn_def_message_ref: OdxLinkRef | None = None
    read_dyn_def_message_snref: str | None = None

    dyn_def_message_ref: OdxLinkRef | None = None
    dyn_def_message_snref: str | None = None

    supported_dyn_ids: list[bytes] = field(default_factory=list)
    selection_table_refs: list[OdxLinkRef | str] = field(default_factory=list)

    @property
    def clear_dyn_def_message(self) -> DiagComm:
        return self._clear_dyn_def_message

    @property
    def read_dyn_def_message(self) -> DiagComm:
        return self._read_dyn_def_message

    @property
    def dyn_def_message(self) -> DiagComm:
        return self._dyn_def_message

    @property
    def selection_tables(self) -> NamedItemList[Table]:
        return self._selection_tables

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "DynIdDefModeInfo":
        def_mode = odxrequire(et_element.findtext("DEF-MODE"))

        clear_dyn_def_message_ref = OdxLinkRef.from_et(
            et_element.find("CLEAR-DYN-DEF-MESSAGE-REF"), context)
        clear_dyn_def_message_snref = None
        if (snref_elem := et_element.find("CLEAR-DYN-DEF-MESSAGE-SNREF")) is not None:
            clear_dyn_def_message_snref = odxrequire(snref_elem.attrib.get("SHORT-NAME"))

        read_dyn_def_message_ref = OdxLinkRef.from_et(
            et_element.find("READ-DYN-DEF-MESSAGE-REF"), context)
        read_dyn_def_message_snref = None
        if (snref_elem := et_element.find("READ-DYN-DEF-MESSAGE-SNREF")) is not None:
            read_dyn_def_message_snref = odxrequire(snref_elem.attrib.get("SHORT-NAME"))

        dyn_def_message_ref = OdxLinkRef.from_et(et_element.find("DYN-DEF-MESSAGE-REF"), context)
        dyn_def_message_snref = None
        if (snref_elem := et_element.find("DYN-DEF-MESSAGE-SNREF")) is not None:
            dyn_def_message_snref = odxrequire(snref_elem.attrib.get("SHORT-NAME"))

        supported_dyn_ids = [
            bytes.fromhex(odxrequire(x.text))
            for x in et_element.iterfind("SUPPORTED-DYN-IDS/SUPPORTED-DYN-ID")
        ]

        selection_table_refs: list[OdxLinkRef | str] = []
        if (st_elems := et_element.find("SELECTION-TABLE-REFS")) is not None:
            for st_elem in st_elems:
                if st_elem.tag == "SELECTION-TABLE-REF":
                    selection_table_refs.append(OdxLinkRef.from_et(st_elem, context))
                elif st_elem.tag == "SELECTION-TABLE-SNREF":
                    selection_table_refs.append(odxrequire(st_elem.get("SHORT-NAME")))
                else:
                    odxraise()

        return DynIdDefModeInfo(
            def_mode=def_mode,
            clear_dyn_def_message_ref=clear_dyn_def_message_ref,
            clear_dyn_def_message_snref=clear_dyn_def_message_snref,
            read_dyn_def_message_ref=read_dyn_def_message_ref,
            read_dyn_def_message_snref=read_dyn_def_message_snref,
            dyn_def_message_ref=dyn_def_message_ref,
            dyn_def_message_snref=dyn_def_message_snref,
            supported_dyn_ids=supported_dyn_ids,
            selection_table_refs=selection_table_refs,
        )

    def __post_init__(self) -> None:
        odxassert(
            self.clear_dyn_def_message_ref is not None or
            self.clear_dyn_def_message_snref is not None,
            "A CLEAR-DYN-DEF-MESSAGE must be specified")
        odxassert(
            self.read_dyn_def_message_ref is not None or
            self.read_dyn_def_message_snref is not None, "A READ-DYN-DEF-MESSAGE must be specified")
        odxassert(self.dyn_def_message_ref is not None or self.dyn_def_message_snref is not None,
                  "A DYN-DEF-MESSAGE must be specified")

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        result: dict[OdxLinkId, Any] = {}

        return result

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        if self.clear_dyn_def_message_ref is not None:
            self._clear_dyn_def_message = odxlinks.resolve(self.clear_dyn_def_message_ref, DiagComm)

        if self.read_dyn_def_message_ref is not None:
            self._read_dyn_def_message = odxlinks.resolve(self.read_dyn_def_message_ref, DiagComm)

        if self.dyn_def_message_ref is not None:
            self._dyn_def_message = odxlinks.resolve(self.dyn_def_message_ref, DiagComm)

        # resolve the selection tables that are referenced via ODXLINK
        self._selection_tables = NamedItemList[Table]()
        for x in self.selection_table_refs:
            if isinstance(x, OdxLinkRef):
                self._selection_tables.append(odxlinks.resolve(x, Table))

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        diag_layer = odxrequire(context.diag_layer)

        if self.clear_dyn_def_message_snref is not None:
            self._clear_dyn_def_message = resolve_snref(self.clear_dyn_def_message_snref,
                                                        diag_layer.diag_comms, DiagComm)

        if self.read_dyn_def_message_snref is not None:
            self._read_dyn_def_message = resolve_snref(self.read_dyn_def_message_snref,
                                                       diag_layer.diag_comms, DiagComm)

        if self.dyn_def_message_snref is not None:
            self._dyn_def_message = resolve_snref(self.dyn_def_message_snref, diag_layer.diag_comms,
                                                  DiagComm)

        if self._clear_dyn_def_message.diagnostic_class != DiagClassType.CLEAR_DYN_DEF_MESSAGE:
            odxraise(
                f"Diagnostic communication object of wrong type referenced: "
                f"({odxrequire(self._clear_dyn_def_message.diagnostic_class).value} instead of "
                f"CLEAR-DYN-DEF-MESSAGE)")
        if self._read_dyn_def_message.diagnostic_class != DiagClassType.READ_DYN_DEFINED_MESSAGE:
            odxraise(f"Diagnostic communication object of wrong type referenced: "
                     f"({odxrequire(self._read_dyn_def_message.diagnostic_class).value} instead of "
                     f"READ-DYN-DEFINED-MESSAGE)")
        if self._dyn_def_message.diagnostic_class != DiagClassType.DYN_DEF_MESSAGE:
            odxraise(f"Diagnostic communication object of wrong type referenced: "
                     f"({odxrequire(self._dyn_def_message.diagnostic_class).value} instead of "
                     f"DYN-DEF-MESSAGE)")

        # resolve the remaining selection tables that are referenced via SNREF
        ddd_spec = odxrequire(diag_layer.diag_data_dictionary_spec)
        for i, x in enumerate(self.selection_table_refs):
            if isinstance(x, str):
                self._selection_tables.insert(i, resolve_snref(x, ddd_spec.tables, Table))



# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any
from xml.etree import ElementTree

from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId
from .snrefcontext import SnRefContext

if TYPE_CHECKING:
    from .matchingparameter import MatchingParameter


@dataclass(kw_only=True)
class SubComponentPattern:
    matching_parameters: list["MatchingParameter"] = field(default_factory=list)

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "SubComponentPattern":
        from .matchingparameter import MatchingParameter

        matching_parameters = [
            MatchingParameter.from_et(el, context)
            for el in et_element.iterfind("MATCHING-PARAMETERS/MATCHING-PARAMETER")
        ]

        return SubComponentPattern(matching_parameters=matching_parameters)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        result = {}
        for mp in self.matching_parameters:
            result.update(mp._build_odxlinks())

        return result

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        for mp in self.matching_parameters:
            mp._resolve_odxlinks(odxlinks)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        for mp in self.matching_parameters:
            mp._resolve_snrefs(context)



# SPDX-License-Identifier: MIT
import unittest
from copy import copy

from examples import somersaultecu
from odxtools.compumethods.compucategory import CompuCategory
from odxtools.compumethods.identicalcompumethod import IdenticalCompuMethod
from odxtools.compumethods.limit import Limit
from odxtools.dataobjectproperty import DataObjectProperty
from odxtools.decodestate import DecodeState
from odxtools.diagdatadictionaryspec import DiagDataDictionarySpec
from odxtools.diagnostictroublecode import DiagnosticTroubleCode
from odxtools.dtcdop import DtcDop
from odxtools.encodestate import EncodeState
from odxtools.environmentdata import EnvironmentData
from odxtools.environmentdatadescription import EnvironmentDataDescription
from odxtools.multiplexer import Multiplexer
from odxtools.multiplexercase import MultiplexerCase
from odxtools.multiplexerdefaultcase import MultiplexerDefaultCase
from odxtools.multiplexerswitchkey import MultiplexerSwitchKey
from odxtools.nameditemlist import NamedItemList
from odxtools.odxlink import DocType, OdxDocFragment, OdxLinkDatabase, OdxLinkId, OdxLinkRef
from odxtools.odxtypes import DataType
from odxtools.parameters.physicalconstantparameter import PhysicalConstantParameter
from odxtools.parameters.valueparameter import ValueParameter
from odxtools.physicaltype import PhysicalType
from odxtools.standardlengthtype import StandardLengthType
from odxtools.structure import Structure
from odxtools.table import Table
from odxtools.tablerow import TableRow
from odxtools.text import Text

# the document fragment which is used throughout the test
doc_frags = [OdxDocFragment("DiagDataDictionarySpecTest", DocType.CONTAINER)]


class TestDiagDataDictionarySpec(unittest.TestCase):

    def test_initialization(self) -> None:
        uint8_dct = StandardLengthType(
            base_data_type=DataType.A_UINT32,
            bit_length=8,
        )
        ident_compu_method = IdenticalCompuMethod(
            category=CompuCategory.IDENTICAL,
            internal_type=DataType.A_UINT32,
            physical_type=DataType.A_UINT32)

        dtc_dop = DtcDop(
            odx_id=OdxLinkId("test_ddds.dop.dtc_dop", doc_frags),
            short_name="dtc_dop",
            diag_coded_type=uint8_dct,
            physical_type=PhysicalType(base_data_type=DataType.A_UINT32),
            compu_method=ident_compu_method,
            dtcs_raw=[
                DiagnosticTroubleCode(
                    odx_id=OdxLinkId("test_ddds.dop.dtc_dop.DTC.X10", doc_frags),
                    short_name="X10",
                    trouble_code=0x10,
                    text=Text.from_string("Something exploded."),
                    display_trouble_code="X10",
                )
            ],
        )

        dop_1 = DataObjectProperty(
            odx_id=OdxLinkId("test_ddds.dop.the_dop", doc_frags),
            short_name="the_dop",
            diag_coded_type=uint8_dct,
            physical_type=PhysicalType(base_data_type=DataType.A_UINT32),
            compu_method=ident_compu_method,
        )

        dop_2 = DataObjectProperty(
            odx_id=OdxLinkId("test_ddds.dop.another_dop", doc_frags),
            short_name="another_dop",
            diag_coded_type=uint8_dct,
            physical_type=PhysicalType(base_data_type=DataType.A_UINT32),
            compu_method=ident_compu_method,
        )

        flip_quality_id = OdxLinkId("test_ddds.table.flip_quality", doc_frags)
        flip_quality_ref = OdxLinkRef.from_id(flip_quality_id)
        table = Table(
            odx_id=flip_quality_id,
            short_name="flip_quality",
            long_name="Flip Quality",
            table_rows_raw=[
                TableRow(
                    odx_id=OdxLinkId("test_ddds.table.flip_quality.average", doc_frags),
                    table_ref=flip_quality_ref,
                    short_name="average",
                    long_name="Average",
                    key_raw="3",
                ),
                TableRow(
                    odx_id=OdxLinkId("test_ddds.table.flip_quality.good", doc_frags),
                    table_ref=flip_quality_ref,
                    short_name="good",
                    long_name="Good",
                    key_raw="5",
                ),
                TableRow(
                    odx_id=OdxLinkId("test_ddds.table.flip_quality.best", doc_frags),
                    table_ref=flip_quality_ref,
                    short_name="best",
                    long_name="Best",
                    key_raw="10",
                ),
            ],
        )

        env_data = EnvironmentData(
            odx_id=OdxLinkId("test_ddds.env_data.flip_env_data", doc_frags),
            short_name="flip_env_data",
            long_name="Flip Env Data",
            parameters=NamedItemList([
                ValueParameter(
                    short_name="flip_speed",
                    long_name="Flip Speed",
                    semantic="DATA",
                    byte_position=0,
                    dop_ref=OdxLinkRef.from_id(dop_1.odx_id),
                ),
                PhysicalConstantParameter(
                    short_name="flip_direction",
                    long_name="Flip Direction",
                    semantic="DATA",
                    byte_position=1,
                    physical_constant_value_raw="1",
                    dop_ref=OdxLinkRef.from_id(dop_1.odx_id),
                ),
            ]),
        )

        env_data_desc = EnvironmentDataDescription(
            odx_id=OdxLinkId("test_ddds.env_data_desc.flip_env_data_desc", doc_frags),
            short_name="flip_env_data_desc",
            long_name="Flip Env Data Desc",
            param_snref="flip_speed",
            env_data_refs=[OdxLinkRef.from_id(env_data.odx_id)],
        )

        mux_case1_struct = Structure(
            odx_id=OdxLinkId("ddds_test.mux.case1.struct", doc_frags),
            short_name="mux_case1_struct",
            parameters=NamedItemList([
                ValueParameter(
                    short_name="min_donation",
                    dop_ref=OdxLinkRef.from_id(dop_1.odx_id),
                ),
                ValueParameter(
                    short_name="min_surface_softness",
                    dop_ref=OdxLinkRef.from_id(dop_1.odx_id),
                )
            ]),
        )

        mux_case2_struct = Structure(
            odx_id=OdxLinkId("ddds_test.mux.case2.struct", doc_frags),
            short_name="mux_case2_struct",
            parameters=NamedItemList([
                ValueParameter(
                    short_name="min_temperature",
                    dop_ref=OdxLinkRef.from_id(dop_1.odx_id),
                ),
            ]),
        )

        mux = Multiplexer(
            odx_id=OdxLinkId("test_ddds.multiplexer.flip_precondition", doc_frags),
            short_name="flip_precondition",
            long_name="Preconditions for doing flips",
            byte_position=1,
            switch_key=MultiplexerSwitchKey(
                byte_position=0,
                bit_position=0,
                dop_ref=OdxLinkRef.from_id(dop_1.odx_id),
            ),
            default_case=MultiplexerDefaultCase(
                short_name="default_case",
                long_name="Preconditions for Other Actions",
            ),
            cases=NamedItemList([
                MultiplexerCase(
                    short_name="forward_flip",
                    long_name="Preconditions for doing a Forward Flip",
                    lower_limit=Limit(value_raw="1", value_type=DataType.A_INT32),
                    upper_limit=Limit(value_raw="3", value_type=DataType.A_INT32),
                    structure_ref=OdxLinkRef.from_id(mux_case1_struct.odx_id),
                ),
                MultiplexerCase(
                    short_name="backward_flip",
                    long_name="Preconditions for doing a Backward Flip",
                    lower_limit=Limit(value_raw="5", value_type=DataType.A_INT32),
                    upper_limit=Limit(value_raw="10", value_type=DataType.A_INT32),
                    structure_ref=OdxLinkRef.from_id(mux_case2_struct.odx_id),
                ),
            ]),
        )

        ddds = DiagDataDictionarySpec(
            dtc_dops=NamedItemList([dtc_dop]),
            data_object_props=NamedItemList([dop_1, dop_2]),
            tables=NamedItemList([table]),
            env_data_descs=NamedItemList([env_data_desc]),
            env_datas=NamedItemList([env_data]),
            muxs=NamedItemList([mux]),
            structures=NamedItemList([mux_case1_struct, mux_case2_struct]),
        )

        # test the short name resolution of TableKeyParameter.
        ecu = somersaultecu.database.ecus.somersault_lazy
        ecu.diag_layer_raw.diag_data_dictionary_spec = ddds
        odxlinks = OdxLinkDatabase()
        odxlinks.update(somersaultecu.database._build_odxlinks())
        odxlinks.update(ecu._build_odxlinks())
        ecu._resolve_odxlinks(odxlinks)
        ecu._finalize_init(database=somersaultecu.database, odxlinks=odxlinks)

        self.assertEqual(ddds.dtc_dops[0], dtc_dop)
        self.assertEqual(ddds.data_object_props[0], dop_1)
        self.assertEqual(ddds.data_object_props.another_dop, dop_2)
        self.assertEqual(ddds.tables[0], table)
        self.assertEqual(ddds.env_data_descs[0], env_data_desc)
        self.assertEqual(ddds.env_datas[0], env_data)
        self.assertEqual(ddds.muxs[0], mux)

        self.assertEqual(len(ddds.structures), 2)
        self.assertEqual(len(ddds.end_of_pdu_fields), 0)

        ddds = ecu.diag_data_dictionary_spec

        # make sure that all locally defined DOPs are present in the finalized ECU
        all_dops = {x.short_name for x in ddds.all_data_object_properties}
        local_dops = {
            "num_flips", "soberness_check", "dizzyness_level", "happiness_level", "duration",
            "temperature", "error_code", "boolean", "uint8", "uint16", "float",
            "flip_env_data_desc", "flip_env_data"
        }
        self.assertEqual(local_dops - all_dops, set())

        # forward flip case
        encode_state = EncodeState(coded_message=bytearray.fromhex("ffee"), cursor_byte_position=2)
        mux.encode_into_pdu(("forward_flip", {
            "min_donation": 3,
            "min_surface_softness": 123
        }), encode_state)
        self.assertEqual(encode_state.coded_message.hex(), "ffee01037b")

        decode_state = DecodeState(
            coded_message=copy(encode_state.coded_message), cursor_byte_position=2)
        decoded = mux.decode_from_pdu(decode_state)
        self.assertEqual(decoded, ("forward_flip", {
            "min_donation": 3,
            "min_surface_softness": 123
        }))

        # backward flip case
        encode_state = EncodeState(coded_message=bytearray.fromhex("ffee"), cursor_byte_position=2)
        mux.encode_into_pdu({7: {"min_temperature": 21}}, encode_state)
        self.assertEqual(encode_state.coded_message.hex(), "ffee0715")

        decode_state = DecodeState(
            coded_message=copy(encode_state.coded_message), cursor_byte_position=2)
        decoded = mux.decode_from_pdu(decode_state)
        self.assertEqual(decoded, ("backward_flip", {"min_temperature": 21}))

        # default case
        encode_state = EncodeState(coded_message=bytearray.fromhex("ffee"), cursor_byte_position=2)
        mux.encode_into_pdu([100, {}], encode_state)
        self.assertEqual(encode_state.coded_message.hex(), "ffee64")

        decode_state = DecodeState(
            coded_message=copy(encode_state.coded_message), cursor_byte_position=2)
        decoded = mux.decode_from_pdu(decode_state)
        self.assertEqual(decoded, ("default_case", {}))


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: MIT
import argparse
import re

from ..database import Database
from ..diagservice import DiagService
from ..exceptions import odxraise
from ..odxtypes import ParameterValue
from ..singleecujob import SingleEcuJob
from . import _parser_utils
from ._parser_utils import SubparsersList

# name of the tool
_odxtools_tool_name_ = "decode"


def get_display_value(v: ParameterValue) -> str:
    if isinstance(v, bytes):
        return v.hex(" ")
    elif isinstance(v, int):
        return f"{v} (0x{v:x})"
    else:
        return str(v)


def print_summary(
    odxdb: Database,
    ecu_variants: list[str] | None = None,
    data: bytes = b'',
    decode: bool = False,
) -> None:
    ecu_names = ecu_variants if ecu_variants else [ecu.short_name for ecu in odxdb.ecus]
    service_db: dict[str, DiagService] = {}
    service_ecus: dict[str, list[str]] = {}
    for ecu_name in ecu_names:
        ecu = odxdb.ecus[ecu_name]
        if not ecu:
            print(f"The ecu variant '{ecu_name}' could not be found!")
            continue
        if data:
            found_services = ecu._find_services_for_uds(data)
            for found_service in found_services:
                ecu_names = service_ecus.get(found_service.short_name, [])
                ecu_names.append(ecu_name)
                service_ecus[found_service.short_name] = ecu_names
                service_db[found_service.short_name] = found_service

    print(f"Binary data: {data.hex(' ')}")
    for service_name, ecu_names in service_ecus.items():
        service = service_db[service_name]
        if isinstance(service, DiagService):
            print(
                f"Decoded by service '{service.short_name}' (decoding ECUs: {', '.join(ecu_names)})"
            )
        elif isinstance(service, SingleEcuJob):
            print(
                f"Decoded by single ecu job '{service.short_name}' (decoding ECUs: {', '.join(ecu_names)})"
            )
        else:
            print(f"Decoded by unknown diagnostic communication: '{service.short_name}' "
                  f"(decoding ECUs: {', '.join(ecu_names)})")

        if decode:
            if data is None:
                odxraise("Data is required for decoding")

            decoded = service.decode_message(data)
            print(f"Decoded data:")
            for param_name, param_value in decoded.param_dict.items():
                print(f"  {param_name}={get_display_value(param_value)}")


def add_subparser(subparsers: SubparsersList) -> None:
    parser = subparsers.add_parser(
        "decode",
        description="\n".join([
            "Decode request by hex-data",
            "",
            "Examples:",
            "  For displaying the service associated with the request 10 01 & decoding it:",
            "    odxtools decode ./path/to/database.pdx -D -d '10 01'",
            "  For displaying the service associated with the request 10 01, without decoding it:",
            "    odxtools decode ./path/to/database.pdx -d '10 01'",
            "  For more information use:",
            "    odxtools decode -h",
        ]),
        help="Find & print service by hex-data. Can also decode the hex-data to its named parameters.",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    _parser_utils.add_pdx_argument(parser)

    parser.add_argument(
        "-v",
        "--variants",
        nargs=1,
        metavar="VARIANT",
        required=False,
        help="Specifies which ecu variants should be included.",
        default="all",
    )

    parser.add_argument(
        "-d",
        "--data",
        metavar="DATA",
        required=True,
        help="Specify data of hex request",
    )

    parser.add_argument(
        "-D",
        "--decode",
        action="store_true",
        required=False,
        help="Decode the given hex data",
    )


def hex_to_binary(data_str: str) -> bytes:
    return bytes.fromhex(data_str)


def run(args: argparse.Namespace) -> None:
    data = bytes.fromhex(re.sub('[^0-9a-fA-F]', '', args.data))
    odxdb = _parser_utils.load_file(args)
    variants = args.variants

    print_summary(
        odxdb,
        ecu_variants=None if variants == "all" else variants,
        data=data,
        decode=args.decode,
    )



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from xml.etree import ElementTree

from .basicstructure import BasicStructure
from .complexdop import ComplexDop
from .environmentdatadescription import EnvironmentDataDescription
from .exceptions import odxassert, odxrequire
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkRef, resolve_snref
from .odxtypes import odxstr_to_bool
from .snrefcontext import SnRefContext
from .utils import dataclass_fields_asdict


@dataclass(kw_only=True)
class Field(ComplexDop):
    structure_ref: OdxLinkRef | None = None
    structure_snref: str | None = None
    env_data_desc_ref: OdxLinkRef | None = None
    env_data_desc_snref: str | None = None
    is_visible_raw: bool | None = None

    @property
    def structure(self) -> BasicStructure:
        """may be a Structure or a env-data-desc"""
        return odxrequire(self._structure, "EnvironmentDataDescription is not supported")

    @property
    def is_visible(self) -> bool:
        return self.is_visible_raw in (None, True)

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "Field":
        kwargs = dataclass_fields_asdict(ComplexDop.from_et(et_element, context))

        structure_ref = OdxLinkRef.from_et(et_element.find("BASIC-STRUCTURE-REF"), context)
        structure_snref = None
        if (edsnr_elem := et_element.find("BASIC-STRUCTURE-SNREF")) is not None:
            structure_snref = edsnr_elem.get("SHORT-NAME")

        env_data_desc_ref = OdxLinkRef.from_et(et_element.find("ENV-DATA-DESC-REF"), context)
        env_data_desc_snref = None
        if (edsnr_elem := et_element.find("ENV-DATA-DESC-SNREF")) is not None:
            env_data_desc_snref = edsnr_elem.get("SHORT-NAME")

        is_visible_raw = odxstr_to_bool(et_element.get("IS-VISIBLE"))
        return Field(
            structure_ref=structure_ref,
            structure_snref=structure_snref,
            env_data_desc_ref=env_data_desc_ref,
            env_data_desc_snref=env_data_desc_snref,
            is_visible_raw=is_visible_raw,
            **kwargs)

    def __post_init__(self) -> None:
        self._structure: BasicStructure | None = None
        self._env_data_desc: EnvironmentDataDescription | None = None
        num_struct_refs = 0 if self.structure_ref is None else 1
        num_struct_refs += 0 if self.structure_snref is None else 1

        num_edd_refs = 0 if self.env_data_desc_ref is None else 1
        num_edd_refs += 0 if self.env_data_desc_snref is None else 1

        odxassert(
            num_struct_refs + num_edd_refs == 1,
            "FIELDs need to specify exactly one reference to a "
            "structure or an environment data description")

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        """Recursively resolve any odxlinks references"""
        if self.structure_ref is not None:
            self._structure = odxlinks.resolve(self.structure_ref, BasicStructure)

        if self.env_data_desc_ref is not None:
            self._env_data_desc = odxlinks.resolve(self.env_data_desc_ref,
                                                   EnvironmentDataDescription)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        """Recursively resolve any short-name references"""
        ddds = odxrequire(context.diag_layer).diag_data_dictionary_spec

        if self.structure_snref is not None:
            self._structure = resolve_snref(self.structure_snref, ddds.structures, BasicStructure)

        if self.env_data_desc_snref is not None:
            self._env_data_desc = resolve_snref(self.env_data_desc_snref, ddds.env_data_descs,
                                                EnvironmentDataDescription)

    def get_static_bit_length(self) -> int | None:
        return None



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any
from xml.etree import ElementTree

from .exceptions import odxassert, odxrequire
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId, OdxLinkRef
from .odxtypes import ParameterValueDict
from .parameters.parameter import Parameter
from .snrefcontext import SnRefContext
from .state import State
from .statetransitionref import _check_applies
from .utils import dataclass_fields_asdict

if TYPE_CHECKING:
    from .statemachine import StateMachine


@dataclass(kw_only=True)
class PreConditionStateRef(OdxLinkRef):
    """
    This class represents the PRE-CONDITION-STATE-REF XML tag.
    """
    value: str | None = None

    in_param_if_snref: str | None = None
    in_param_if_snpathref: str | None = None

    @property
    def state(self) -> "State":
        return self._state

    @staticmethod
    def from_et(  # type: ignore[override]
            et_element: ElementTree.Element, context: OdxDocContext) -> "PreConditionStateRef":
        kwargs = dataclass_fields_asdict(OdxLinkRef.from_et(et_element, context))

        value = et_element.findtext("VALUE")

        in_param_if_snref = None
        if (in_param_if_snref_elem := et_element.find("IN-PARAM-IF-SNREF")) is not None:
            in_param_if_snref = odxrequire(in_param_if_snref_elem.get("SHORT-NAME"))

        in_param_if_snpathref = None
        if (in_param_if_snpathref_elem := et_element.find("IN-PARAM-IF-SNPATHREF")) is not None:
            in_param_if_snpathref = odxrequire(in_param_if_snpathref_elem.get("SHORT-NAME-PATH"))

        return PreConditionStateRef(
            value=value,
            in_param_if_snref=in_param_if_snref,
            in_param_if_snpathref=in_param_if_snpathref,
            **kwargs)

    def __post_init__(self) -> None:
        if self.value is not None:
            odxassert(self.in_param_if_snref is not None or self.in_param_if_snref is not None,
                      "If VALUE is specified, a parameter must be referenced")

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        return {}

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        self._state = odxlinks.resolve(self, State)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        pass

    def applies(self, state_machine: "StateMachine", params: list[Parameter],
                param_value_dict: ParameterValueDict) -> bool:
        """Given a state machine, evaluate whether the precondition is fulfilled or not

        Note that the specification is unclear about what the
        parameters are: It says "The optional VALUE together with the
        also optional IN-PARAM-IF snref at STATE-TRANSITION-REF and
        PRE-CONDITION-STATE-REF can be used if the STATE-TRANSITIONs
        and pre-condition STATEs are dependent on the values of the
        referenced PARAMs.", but it does not specify what the
        "referenced PARAMs" are. For the state transition refs, let's
        assume that they are the parameters of a positive response
        received from the ECU, whilst we assume that they are the
        parameters of the request for PRE-CONDITION-STATE-REFs.
        """

        return _check_applies(self, state_machine, params, param_value_dict)



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Union
from xml.etree import ElementTree

from .basicstructure import BasicStructure
from .dataobjectproperty import DataObjectProperty
from .exceptions import odxassert, odxraise, odxrequire
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId, OdxLinkRef, resolve_snref
from .odxtypes import ParameterValue, ParameterValueDict
from .parameters.codedconstparameter import CodedConstParameter
from .parameters.parameter import Parameter
from .parameters.physicalconstantparameter import PhysicalConstantParameter
from .parameters.tablekeyparameter import TableKeyParameter
from .parameters.tablestructparameter import TableStructParameter
from .parameters.valueparameter import ValueParameter
from .snrefcontext import SnRefContext
from .state import State
from .statemachine import StateMachine
from .statetransition import StateTransition
from .utils import dataclass_fields_asdict

if TYPE_CHECKING:
    from .preconditionstateref import PreConditionStateRef
    from .statemachine import StateMachine
    from .tablerow import TableRow


def _resolve_in_param(
    in_param_if_snref: str | None,
    in_param_if_snpathref: str | None,
    params: list[Parameter],
    param_dict: ParameterValueDict,
) -> tuple[Parameter | None, ParameterValue | None]:

    if in_param_if_snref is not None:
        path_chunks = [in_param_if_snref]
    elif in_param_if_snpathref is not None:
        path_chunks = in_param_if_snpathref.split(".")
    else:
        return None, None

    return _resolve_in_param_helper(params, param_dict, path_chunks)


def _resolve_in_param_helper(
    params: list[Parameter],
    param_dict: ParameterValueDict,
    path_chunks: list[str],
) -> tuple[Parameter | None, ParameterValue | None]:

    inner_param = resolve_snref(path_chunks[0], params, Parameter, lenient=True)
    if inner_param is None:
        return None, None
    inner_param_value = param_dict.get(path_chunks[0])

    if len(path_chunks) == 1:
        return inner_param, inner_param_value

    # deal with table parameters
    if isinstance(inner_param, TableStructParameter):
        if not isinstance(inner_param_value, tuple) or len(inner_param_value) == 2:
            odxraise("Invalid value for table struct parameter")
            return None, None

        if TYPE_CHECKING:
            tr = resolve_snref(inner_param_value[0], inner_param.table.table_rows, TableRow)
        else:
            tr = resolve_snref(inner_param_value[0], inner_param.table.table_rows)

        if tr.structure is None:
            odxraise("Table row must reference a structure")
            return None, None

        inner_param_list = tr.structure.parameters
        inner_param_value = inner_param_value[1]
    else:
        dop = getattr(inner_param, "dop", None)
        if isinstance(dop, BasicStructure):
            inner_param_list = dop.parameters
        else:
            odxraise(f"Expected the referenced parameter to be composite")

    if not isinstance(inner_param_value, dict):
        odxraise(f"Expected the referenced parameter to be composite")
        return None, None

    return _resolve_in_param_helper(inner_param_list, inner_param_value, path_chunks[1:])


def _check_applies(ref: Union["StateTransitionRef",
                              "PreConditionStateRef"], state_machine: "StateMachine",
                   params: list[Parameter], param_value_dict: ParameterValueDict) -> bool:
    if state_machine.active_state != ref.state:
        # if the active state of the state machine is not the
        # specified one, the precondition does not apply
        return False

    if ref.in_param_if_snref is None and ref.in_param_if_snpathref is None:
        # no parameter given -> only the specified state is relevant
        return True

    param, param_value = \
        _resolve_in_param(ref.in_param_if_snref, ref.in_param_if_snpathref, params, param_value_dict)

    if param is None:
        # The referenced parameter does not exist.
        return False
    elif not isinstance(
            param,
        (CodedConstParameter, PhysicalConstantParameter, TableKeyParameter, ValueParameter)):
        # see checker rule 194 in section B.2 of the spec
        odxraise(f"Parameter referenced by state transition ref is of "
                 f"invalid type {type(param).__name__}")
        return False
    elif isinstance(param, (CodedConstParameter, PhysicalConstantParameter,
                            TableKeyParameter)) and ref.value is not None:
        # see checker rule 193 in section B.2 of the spec. Why can
        # no values for constant parameters be specified? (This
        # seems to be rather inconvenient...)
        odxraise(f"No value may be specified for state transitions referencing "
                 f"parameters of type {type(param).__name__}")
        return False
    elif isinstance(param, ValueParameter):
        # see checker rule 193 in section B.2 of the spec
        if ref.value is None:
            odxraise(f"An expected VALUE must be specified for preconditions "
                     f"referencing VALUE parameters")
            return False

        if param_value is None:
            param_value = param.physical_default_value
            if param_value is None:
                odxraise(f"No value can be determined for parameter {param.short_name}")
                return False

        if param.dop is None or not isinstance(param.dop, DataObjectProperty):
            odxraise(f"Parameter {param.short_name} uses a non-simple DOP")
            return False

        expected_value = param.dop.physical_type.base_data_type.from_string(ref.value)

        if expected_value != param_value:
            return False

    return True


@dataclass(kw_only=True)
class StateTransitionRef(OdxLinkRef):
    """Describes a state transition that is to be potentially taken if
    a diagnostic communication is executed

    Besides the "raw" state transistion, state transition references
    may also be conditional on the observed response of the ECU.

    """
    value: str | None = None

    in_param_if_snref: str | None = None
    in_param_if_snpathref: str | None = None

    @property
    def state_transition(self) -> StateTransition:
        return self._state_transition

    @property
    def state(self) -> State:
        return self.state_transition.source_state

    @staticmethod
    def from_et(  # type: ignore[override]
            et_element: ElementTree.Element, context: OdxDocContext) -> "StateTransitionRef":
        kwargs = dataclass_fields_asdict(OdxLinkRef.from_et(et_element, context))

        value = et_element.findtext("VALUE")

        in_param_if_snref = None
        if (in_param_if_snref_elem := et_element.find("IN-PARAM-IF-SNREF")) is not None:
            in_param_if_snref = odxrequire(in_param_if_snref_elem.get("SHORT-NAME"))

        in_param_if_snpathref = None
        if (in_param_if_snpathref_elem := et_element.find("IN-PARAM-IF-SNPATHREF")) is not None:
            in_param_if_snpathref = odxrequire(in_param_if_snpathref_elem.get("SHORT-NAME-PATH"))

        return StateTransitionRef(
            value=value,
            in_param_if_snref=in_param_if_snref,
            in_param_if_snpathref=in_param_if_snpathref,
            **kwargs)

    def __post_init__(self) -> None:
        if self.value is not None:
            odxassert(self.in_param_if_snref is not None or self.in_param_if_snref is not None,
                      "If VALUE is specified, a parameter must be referenced")

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        return {}

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        self._state_transition = odxlinks.resolve(self, StateTransition)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        pass

    def execute(self, state_machine: StateMachine, params: list[Parameter],
                param_value_dict: ParameterValueDict) -> bool:
        """Update a StateMachine object if the state transition ought
        to be executed based on the response received after executing a
        diagnostic communication

        Note that the specification is unclear about what the
        parameters are: It says "The optional VALUE together with the
        also optional IN-PARAM-IF snref at STATE-TRANSITION-REF and
        PRE-CONDITION-STATE-REF can be used if the STATE-TRANSITIONs
        and pre-condition STATEs are dependent on the values of the
        referenced PARAMs.", but it does not specify what the
        "referenced PARAMs" are. For the state transition refs, let's
        assume that they are the parameters of a response received
        from the ECU, whilst we assume that they are the parameters of
        the request for PRE-CONDITION-STATE-REFs.

        Returns:
            ``True´` if the state transition was taken, else ``False``
        """
        if _check_applies(self, state_machine, params, param_value_dict):
            state_machine.active_state = self.state_transition.target_state
            return True

        return False



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import Any
from xml.etree import ElementTree

from .element import NamedElement
from .exceptions import odxrequire
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId, OdxLinkRef, resolve_snref
from .snrefcontext import SnRefContext
from .table import Table
from .tablerow import TableRow
from .utils import dataclass_fields_asdict


@dataclass(kw_only=True)
class TableRowConnector(NamedElement):
    table_ref: OdxLinkRef
    table_row_snref: str

    @property
    def table(self) -> Table:
        return self._table

    @property
    def table_row(self) -> TableRow:
        return self._table_row

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "TableRowConnector":
        kwargs = dataclass_fields_asdict(NamedElement.from_et(et_element, context))

        table_ref = odxrequire(OdxLinkRef.from_et(et_element.find("TABLE-REF"), context))
        table_row_snref_el = odxrequire(et_element.find("TABLE-ROW-SNREF"))
        table_row_snref = odxrequire(table_row_snref_el.get("SHORT-NAME"))

        return TableRowConnector(table_ref=table_ref, table_row_snref=table_row_snref, **kwargs)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        return {}

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        self._table = odxlinks.resolve(self.table_ref, Table)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        self._table_row = resolve_snref(self.table_row_snref, self._table.table_rows, TableRow)



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import Any
from xml.etree import ElementTree

from .admindata import AdminData
from .element import IdentifiableElement
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId
from .snrefcontext import SnRefContext
from .utils import dataclass_fields_asdict


@dataclass(kw_only=True)
class FunctionalClass(IdentifiableElement):
    """
    Corresponds to FUNCT-CLASS.
    """

    admin_data: AdminData | None = None

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "FunctionalClass":

        kwargs = dataclass_fields_asdict(IdentifiableElement.from_et(et_element, context))

        admin_data = AdminData.from_et(et_element.find("ADMIN-DATA"), context)

        return FunctionalClass(admin_data=admin_data, **kwargs)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        return {self.odx_id: self}

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        pass

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        pass



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import Optional, overload
from xml.etree import ElementTree

from ..exceptions import odxraise
from ..odxdoccontext import OdxDocContext
from ..odxtypes import AtomicOdxType, DataType, compare_odx_values
from .intervaltype import IntervalType


@dataclass(kw_only=True)
class Limit:
    value_raw: str | None = None
    value_type: DataType | None = None
    interval_type: IntervalType | None = None

    @property
    def value(self) -> AtomicOdxType | None:
        return self._value

    @staticmethod
    @overload
    def limit_from_et(et_element: ElementTree.Element, context: OdxDocContext,
                      value_type: DataType | None) -> "Limit":
        ...

    @staticmethod
    @overload
    def limit_from_et(et_element: None, context: OdxDocContext,
                      value_type: DataType | None) -> None:
        ...

    @staticmethod
    def limit_from_et(et_element: ElementTree.Element | None, context: OdxDocContext,
                      value_type: DataType | None) -> Optional["Limit"]:

        if et_element is None:
            return None

        interval_type = None
        if (interval_type_str := et_element.get("INTERVAL-TYPE")) is not None:
            try:
                interval_type = IntervalType(interval_type_str)
            except ValueError:
                odxraise(f"Encountered unknown interval type '{interval_type_str}'")

        value_raw = et_element.text

        return Limit(value_raw=value_raw, interval_type=interval_type, value_type=value_type)

    def __post_init__(self) -> None:
        self._value: AtomicOdxType | None = None

        if self.value_type is not None:
            self.set_value_type(self.value_type)

    def set_value_type(self, value_type: DataType) -> None:
        self.value_type = value_type
        if self.value_raw is not None:
            self._value = value_type.from_string(self.value_raw)

    def complies_to_upper(self, value: AtomicOdxType) -> bool:
        """Checks if the value is in the range w.r.t. the upper limit.

        * If the interval type is closed, return `value <= limit.value`.
        * If the interval type is open, return `value < limit.value`.
        * If the interval type is infinite, return `True`.
        """
        if self._value is None:
            # if no value is specified, assume interval type INFINITE
            # (what are we supposed to compare against?)
            return True

        if self.interval_type is None or self.interval_type == IntervalType.CLOSED:
            # assume interval type CLOSED if a value was specified,
            # but no interval type
            return compare_odx_values(value, self._value) <= 0
        elif self.interval_type == IntervalType.OPEN:
            return compare_odx_values(value, self._value) < 0

        if self.interval_type != IntervalType.INFINITE:
            odxraise("Unhandled interval type {self.interval_type}")

        return True

    def complies_to_lower(self, value: AtomicOdxType) -> bool:
        """Checks if the value is in the range w.r.t. the lower limit.

        * If the interval type is closed, return `limit.value <= value`.
        * If the interval type is open, return `limit.value < value`.
        * If the interval type is infinite, return `True`.
        """

        if self._value is None:
            # if no value is specified, assume interval type INFINITE
            # (what are we supposed to compare against?)
            return True

        if self.interval_type is None or self.interval_type == IntervalType.CLOSED:
            # assume interval type CLOSED if a value was specified,
            # but no interval type
            return compare_odx_values(value, self._value) >= 0
        elif self.interval_type == IntervalType.OPEN:
            return compare_odx_values(value, self._value) > 0

        if self.interval_type != IntervalType.INFINITE:
            odxraise("Unhandled interval type {self.interval_type}")

        return True



# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from xml.etree import ElementTree

from .compumethods.limit import Limit
from .odxdoccontext import OdxDocContext
from .odxtypes import DataType
from .scaleconstr import ScaleConstr


@dataclass(kw_only=True)
class InternalConstr:
    """This class represents INTERNAL-CONSTR objects.
    """

    # TODO: Enforce the internal and physical constraints.

    lower_limit: Limit | None = None
    upper_limit: Limit | None = None
    scale_constrs: list[ScaleConstr] = field(default_factory=list)

    value_type: DataType

    @staticmethod
    def constr_from_et(et_element: ElementTree.Element, context: OdxDocContext, *,
                       value_type: DataType) -> "InternalConstr":

        lower_limit = Limit.limit_from_et(
            et_element.find("LOWER-LIMIT"), context, value_type=value_type)
        upper_limit = Limit.limit_from_et(
            et_element.find("UPPER-LIMIT"), context, value_type=value_type)

        scale_constrs = [
            ScaleConstr.scale_constr_from_et(sc_el, context, value_type=value_type)
            for sc_el in et_element.iterfind("SCALE-CONSTRS/SCALE-CONSTR")
        ]

        return InternalConstr(
            lower_limit=lower_limit,
            upper_limit=upper_limit,
            scale_constrs=scale_constrs,
            value_type=value_type)



# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field

from ..exceptions import odxraise, odxrequire
from ..odxtypes import AtomicOdxType, DataType
from .compuscale import CompuScale
from .limit import Limit


@dataclass(kw_only=True)
class RatFuncSegment:
    """Helper class to represent a segment of a piecewise rational function.
    """
    value_type: DataType

    numerator_coeffs: list[int | float] = field(default_factory=list)
    denominator_coeffs: list[int | float] = field(default_factory=list)

    lower_limit: Limit | None = None
    upper_limit: Limit | None = None

    @staticmethod
    def from_compu_scale(scale: CompuScale, value_type: DataType) -> "RatFuncSegment":
        coeffs = odxrequire(scale.compu_rational_coeffs,
                            "Scales for rational functions must exhibit "
                            "COMPU-RATIONAL-COEFFS")

        numerator_coeffs = coeffs.numerators
        denominator_coeffs = coeffs.denominators

        lower_limit = scale.lower_limit
        upper_limit = scale.upper_limit

        return RatFuncSegment(
            numerator_coeffs=numerator_coeffs,
            denominator_coeffs=denominator_coeffs,
            lower_limit=lower_limit,
            upper_limit=upper_limit,
            value_type=scale.range_type)

    def convert(self, value: AtomicOdxType) -> float | int:
        if not isinstance(value, (int, float)):
            odxraise(f"Internal values of linear compumethods must "
                     f"either be int or float (is: {type(value).__name__})")

        numerator = 0.0
        x = float(value)
        for numerator_coeff in reversed(self.numerator_coeffs):
            numerator *= x
            numerator += float(numerator_coeff)

        denominator = 0.0
        for denominator_coeff in reversed(self.denominator_coeffs):
            denominator *= x
            denominator += float(denominator_coeff)

        result = numerator / denominator

        if self.value_type in [
                DataType.A_INT32,
                DataType.A_UINT32,
        ]:
            result = round(result)

        return result

    def applies(self, value: AtomicOdxType) -> bool:
        """Returns True iff the segment is applicable to a given internal value"""
        # Do type checks
        expected_type = self.value_type.python_type
        if issubclass(expected_type, float):
            if not isinstance(value, (int, float)):
                return False
        else:
            if not isinstance(value, expected_type):
                return False

        if self.lower_limit is not None and \
           not self.lower_limit.complies_to_lower(value):
            return False

        if self.upper_limit is not None and \
           not self.upper_limit.complies_to_upper(value):
            return False

        return True



from dataclasses import dataclass, field
from typing import Optional
from xml.etree import ElementTree

from .exceptions import odxrequire
from .externaldoc import ExternalDoc
from .odxdoccontext import OdxDocContext


@dataclass(kw_only=True)
class Description:
    text: str
    external_docs: list[ExternalDoc] = field(default_factory=list)

    text_identifier: str | None = None

    @staticmethod
    def from_et(et_element: ElementTree.Element | None,
                context: OdxDocContext) -> Optional["Description"]:
        if et_element is None:
            return None

        # Extract the contents of the tag as a XHTML string.
        raw_string = et_element.text or ""
        for e in et_element:
            if e.tag == "EXTERNAL-DOCS":
                break
            raw_string += ElementTree.tostring(e, encoding="unicode")

        # remove white spaces at the beginning and at the end of all
        # extracted lines
        stripped_lines = [x.strip() for x in raw_string.split("\n")]

        text = "\n".join(stripped_lines).strip()

        external_docs = \
            [
                odxrequire(ExternalDoc.from_et(ed, context)) for ed in et_element.iterfind("EXTERNAL-DOCS/EXTERNAL-DOC")
            ]

        text_identifier = et_element.attrib.get("TI")

        return Description(text=text, text_identifier=text_identifier, external_docs=external_docs)

    @staticmethod
    def from_string(text: str) -> "Description":
        return Description(text=text, external_docs=[], text_identifier=None)

    def __str__(self) -> str:
        return self.text



# SPDX-License-Identifier: MIT
"""odxtools is a set of utilities to deal with automotive diagnostic
descriptions of electronic control units using ODX-formatted
description files.

ODX stands for "Open Diagnostic Data Exchange" `[ODX]`_ and is an XML
based file format to describe the diagnostic capabilities of the
various electronic control units (ECUs) of complex distributed
technical systems (mainly cars and trucks). ODX is an open standard,
maintained by ASAM e.V. `[ASAM]`_ and standardized internationally by
ISO 22901 `[ISO22901]`_.

The services provided by odxtools encompass parsing and internalizing
ODX diagnostic database files as well as decoding and endcoding raw
data of diagnostic requests and responses that are send to/received
from ECUs in an pythonic manner.

Examples:
=========

- Load a database from file ``my_diag_db.pdx``:

.. code-block:: python

    import odxtools_porsche

    db = odxtools.load_file("my_diag_db.pdx")

- List the names of all available services of the first ECU:

.. code-block:: python

    # [...]

    dl = db.diag_layer_containers[0].diag_layers[0]
    print(f"Available services for {dl.short_name}: {[s.short_name for s in dl.services]}")

- Determine the CAN IDs which the first described ECU uses to send and receive diagnostic messages:

.. code-block:: python

    # [...]

    print(f"ECU listens on ID 0x{dl.get_receive_id():x}")
    print(f"ECU transmits on ID 0x{dl.get_send_id():x}")

- Decode a diagnostic message for an ECU:

.. code-block:: python

    # [...]

    raw_data = b"\xB0\xA7"
    decoded_message = dl.decode(raw_data)
    print(f"decoded message: {decoded_message}")

References
==========

- _`[ASAM]` ASAM e.V: https://www.asam.net/
- _`[ODX]` The ODX Specification: https://www.asam.net/standards/detail/mcd-2-d/
- _`[ISO22901]` The ISO 22901 Standard: https://www.iso.org/standard/41207.html

"""
from .loadfile import load_directory as load_directory  # noqa: F401
from .loadfile import load_file as load_file  # noqa: F401
from .loadfile import load_files  # noqa: F401
from .loadfile import load_odx_d_file as load_odx_d_file  # noqa: F401
from .loadfile import load_pdx_file as load_pdx_file  # noqa: F401
from .version import __version__ as __version__  # noqa: F401
from .writepdxfile import write_pdx_file as write_pdx_file  # noqa: F401

__author__ = "Katrin Bauer"


def _main() -> None:
    # Command line tool
    from .cli import main as _main

    _main.start_cli()



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import Any, cast
from xml.etree import ElementTree

from .compumethods.compumethod import CompuMethod
from .compumethods.createanycompumethod import create_any_compu_method_from_et
from .createanydiagcodedtype import create_any_diag_coded_type_from_et
from .decodestate import DecodeState
from .diagcodedtype import DiagCodedType
from .dopbase import DopBase
from .encodestate import EncodeState
from .exceptions import EncodeError, odxraise, odxrequire
from .internalconstr import InternalConstr
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId, OdxLinkRef
from .odxtypes import AtomicOdxType, BytesTypes, ParameterValue
from .physicaltype import PhysicalType
from .snrefcontext import SnRefContext
from .unit import Unit
from .utils import dataclass_fields_asdict


@dataclass(kw_only=True)
class DataObjectProperty(DopBase):
    """This class represents a DATA-OBJECT-PROP.

    Note that this class only represents non-complex DOPs. A better
    name would thus be SimpleDataObjectProp...
    """

    #: Conversion from the physical to the internal representation and vice-versa.
    compu_method: CompuMethod

    #: The type used to represent a value internally
    diag_coded_type: DiagCodedType

    #: The type of the value in the physical world
    physical_type: PhysicalType

    internal_constr: InternalConstr | None = None

    #: The unit associated with physical values (e.g. 'm/s^2')
    unit_ref: OdxLinkRef | None = None

    physical_constr: InternalConstr | None = None

    @property
    def unit(self) -> Unit | None:
        return self._unit

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "DataObjectProperty":
        """Reads a DATA-OBJECT-PROP."""
        kwargs = dataclass_fields_asdict(DopBase.from_et(et_element, context))

        diag_coded_type = create_any_diag_coded_type_from_et(
            odxrequire(et_element.find("DIAG-CODED-TYPE")), context)
        physical_type = PhysicalType.from_et(odxrequire(et_element.find("PHYSICAL-TYPE")), context)
        compu_method = create_any_compu_method_from_et(
            odxrequire(et_element.find("COMPU-METHOD")),
            context,
            internal_type=diag_coded_type.base_data_type,
            physical_type=physical_type.base_data_type,
        )
        internal_constr = None
        if (internal_constr_elem := et_element.find("INTERNAL-CONSTR")) is not None:
            internal_constr = InternalConstr.constr_from_et(
                internal_constr_elem, context, value_type=diag_coded_type.base_data_type)
        unit_ref = OdxLinkRef.from_et(et_element.find("UNIT-REF"), context)
        physical_constr = None
        if (physical_constr_elem := et_element.find("PHYS-CONSTR")) is not None:
            physical_constr = InternalConstr.constr_from_et(
                physical_constr_elem, context, value_type=physical_type.base_data_type)

        return DataObjectProperty(
            compu_method=compu_method,
            diag_coded_type=diag_coded_type,
            physical_type=physical_type,
            internal_constr=internal_constr,
            unit_ref=unit_ref,
            physical_constr=physical_constr,
            **kwargs)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        result = super()._build_odxlinks()
        result.update(self.compu_method._build_odxlinks())
        result.update(self.diag_coded_type._build_odxlinks())
        return result

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        """Resolves the reference to the unit"""
        super()._resolve_odxlinks(odxlinks)

        self.compu_method._resolve_odxlinks(odxlinks)
        self.diag_coded_type._resolve_odxlinks(odxlinks)

        self._unit: Unit | None = None
        if self.unit_ref:
            self._unit = odxlinks.resolve(self.unit_ref, Unit)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        super()._resolve_snrefs(context)

        self.compu_method._resolve_snrefs(context)
        self.diag_coded_type._resolve_snrefs(context)

    def get_static_bit_length(self) -> int | None:
        return self.diag_coded_type.get_static_bit_length()

    def encode_into_pdu(self, physical_value: ParameterValue, encode_state: EncodeState) -> None:
        """
        Convert a physical representation of a parameter to a string bytes that can be send over the wire
        """
        if not self.is_valid_physical_value(physical_value):
            raise EncodeError(
                f"The value {repr(physical_value)} of type {type(physical_value).__name__}"
                f" is not a valid.")

        if not isinstance(physical_value, (int, float, str, BytesTypes)):
            odxraise(f"Invalid type '{type(physical_value).__name__}' for physical value. "
                     f"(Expect atomic type!)")
        internal_value = self.compu_method.convert_physical_to_internal(physical_value)
        self.diag_coded_type.encode_into_pdu(internal_value, encode_state)

    def decode_from_pdu(self, decode_state: DecodeState) -> ParameterValue:
        """
        Convert the internal representation of a value into its physical value.

        Returns a (physical_value, start_position_of_next_parameter) tuple.
        """
        internal = self.diag_coded_type.decode_from_pdu(decode_state)

        if self.compu_method.is_valid_internal_value(internal):
            return self.compu_method.convert_internal_to_physical(internal)

        internal_to_phys = self.compu_method.compu_internal_to_phys
        default_value = internal_to_phys.compu_default_value if internal_to_phys else None

        if default_value and default_value.value is not None:
            return default_value.value

        odxraise(f"DOP {self.short_name} could not convert the coded value "
                 f"{repr(internal)} to physical type {self.physical_type.base_data_type}.")

        return None

    def is_valid_physical_value(self, physical_value: ParameterValue) -> bool:
        return self.compu_method.is_valid_physical_value(cast(AtomicOdxType, physical_value))



# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from typing import Any
from xml.etree import ElementTree

from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId
from .relateddoc import RelatedDoc
from .snrefcontext import SnRefContext
from .specialdatagroup import SpecialDataGroup


@dataclass(kw_only=True)
class CompanySpecificInfo:
    related_docs: list[RelatedDoc] = field(default_factory=list)
    sdgs: list[SpecialDataGroup] = field(default_factory=list)

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "CompanySpecificInfo":
        related_docs = [
            RelatedDoc.from_et(rd, context)
            for rd in et_element.iterfind("RELATED-DOCS/RELATED-DOC")
        ]

        sdgs = [SpecialDataGroup.from_et(sdge, context) for sdge in et_element.iterfind("SDGS/SDG")]

        return CompanySpecificInfo(related_docs=related_docs, sdgs=sdgs)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        result = {}

        for rd in self.related_docs:
            result.update(rd._build_odxlinks())

        for sdg in self.sdgs:
            result.update(sdg._build_odxlinks())

        return result

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        for rd in self.related_docs:
            rd._resolve_odxlinks(odxlinks)

        for sdg in self.sdgs:
            sdg._resolve_odxlinks(odxlinks)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        for rd in self.related_docs:
            rd._resolve_snrefs(context)

        for sdg in self.sdgs:
            sdg._resolve_snrefs(context)
